""" @brief  A module for implementing Algs. 1–4 from [E24].

    [E24] Ekerå, M.: "On the success probability of quantum order finding".
                     ACM Trans. Quantum Comput. 5(2):11 (2024). """

from gmpy2 import mpz;

from math import floor;
from math import log;
from math import prod;

from ......math.groups import CyclicGroupElement;

from ......math.primes import prime_range;

from ......utils.timeout import Timeout;

def is_valid_r_tilde(
  r_tilde : int | mpz,
  m : int) -> bool:

  """ @brief  Checks if r_tilde is an integer on [1, 2^m).

      @param r_tilde  The r_tilde to check.

      @param m  The positive integer m.

      @return   True if r_tilde is an integer on [1, 2^m), False otherwise. """

  if not (isinstance(r_tilde, int) or isinstance(r_tilde, mpz)):
    return False;

  return (r_tilde >= 1) and (r_tilde < 2 ** m);


def algorithm1(
  g : CyclicGroupElement,
  r_tilde : int | mpz,
  m : int,
  c : int = 1,
  timeout : int | None | Timeout = None) -> int | mpz | None:

  """ @brief  Recovers a multiple rp of r, assuming r_tilde is such that
              r = d * r_tilde where d is cm-smooth.

      This function implements Alg. 1 from [E24].

      [E24] Ekerå, M.: "On the success probability of quantum order finding".
                       ACM Trans. Quantum Comput. 5(2):11 (2024).

      As in [E24], d is said to be cm-smooth if d = p1^e1 * .. pk^ek, for
      q1, ..., qk pairwise distinct primes, and e1, ..., ek positive integer
      exponents, if it holds that qi^ei <= cm for all i in [1, k].

      @param g  The element g of order r.

      @param r_tilde  The integer r_tilde.

      @param m  A positive integer m such that r < 2^m.

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing cm-smooth component d in r = d * r_tilde.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the computation aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @return   A multiple rp of r, assuming that r_tilde is such that
                r = d * r_tilde where d is cm-smooth, and None otherwise. """

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  # Step 1:
  if not is_valid_r_tilde(r_tilde, m):
    # Step 1.1:
    return None;

  # Step 2:
  rp = r_tilde;
  x = g ** r_tilde;

  # Step 3:
  for q in prime_range(floor(c * m) + 1):

    # Check the timeout.
    timeout.check();

    # Step 3.1:
    if x == 1:
      # Step 3.1.1:
      return rp;

    # Step 3.2:
    e = floor(log(c * m) / log(q));

    # Step 3.3:
    q_pow_e = q ** e;
    x = x ** q_pow_e;
    rp = rp * q_pow_e;

  # Step 4:
  if x != 1:
    # Step 4.1:
    return None;

  # Step 5:
  return rp;


def algorithm2(
  g : CyclicGroupElement,
  r_tilde : int | mpz,
  m : int,
  c : int = 1,
  timeout : int | None | Timeout = None) -> int | mpz | None:

  """ @brief  Recovers r, assuming r_tilde is such that r = d * r_tilde where d
              is cm-smooth.

      This function implements Alg. 2 from [E24].

      [E24] Ekerå, M.: "On the success probability of quantum order finding".
                       ACM Trans. Quantum Comput. 5(2):11 (2024).

      As in [E24], d is said to be cm-smooth if d = p1^e1 * .. pk^ek, for
      q1, ..., qk pairwise distinct primes, and e1, ..., ek positive integer
      exponents, if it holds that qi^ei <= cm for all i in [1, k].

      @param g  The element g of order r.

      @param r_tilde  The integer r_tilde.

      @param m  A positive integer m such that r < 2^m.

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing cm-smooth component d in r = d * r_tilde.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the computation aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @return   The order r, assuming r_tilde is such that r = d * r_tilde where
                d is cm-smooth. Otherwise, None or some positive integer
                multiple of r, is returned. """

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  # Step 1:
  if not is_valid_r_tilde(r_tilde, m):
    # Step 1.1:
    return None;

  # Step 2:
  x = g ** r_tilde;

  # Step 3:
  if x == 1:
    return r_tilde;

  # Step 4:
  S = [];

  # Step 5:
  for q in prime_range(floor(c * m) + 1):

    # Check the timeout.
    timeout.check();

    # Step 5.1:
    e = floor(log(c * m) / log(q));

    # Step 5.2:
    S.append([x, q, e]);

    # Step 5.3:
    q_pow_e = q ** e;
    x = x ** q_pow_e;

    # Step 5.4:
    if x == 1:
      # Step 5.4.1:
      break;

  # Step 6:
  if x != 1:
    # Step 6.1:
    return None;

  # Step 7:
  d = 1;

  # Step 8:
  while S != []:
    [x, q, e] = S.pop();

    # Check the timeout.
    timeout.check();

    # Step 8.1:
    x = x ** d;

    # Step 8.2:
    for i in range(1, e + 1):
      # Step 8.2.1:
      if x == 1:
        # Step 8.2.1.1:
        break;

      # Step 8.2.2:
      x = x ** q;
      d = d * q;

  # Step 9:
  return d * r_tilde;


def algorithm3(
  g : CyclicGroupElement,
  r_tilde : int | mpz,
  m : int,
  c : int = 1,
  timeout : int | None | Timeout = None) -> int | mpz | None:

  """ @brief  Recovers r, assuming r_tilde is such that r = d * r_tilde where d
              is cm-smooth.

      This function implements Alg. 3 from [E24].

      [E24] Ekerå, M.: "On the success probability of quantum order finding".
                       ACM Trans. Quantum Comput. 5(2):11 (2024).

      As in [E24], d is said to be cm-smooth if d = p1^e1 * .. pk^ek, for
      q1, ..., qk pairwise distinct primes, and e1, ..., ek positive integer
      exponents, if it holds that qi^ei <= cm for all i in [1, k].

      @remark   As is explained in [E24], this algorithm is equivalent to Alg. 2
                in that it computes the same result. It has a better worst-case
                time complexity than Alg. 2, but it has a much higher time
                complexity in practice for many problem instances. This explains
                why both algorithms are included.

      @param g  The element g of order r.

      @param r_tilde  The integer r_tilde.

      @param m  A positive integer m such that r < 2^m.

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing cm-smooth component d in r = d * r_tilde.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the computation aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @return   The order r, assuming r_tilde is such that r = d * r_tilde where
                d is cm-smooth. Otherwise, None or some positive integer
                multiple of r, is returned. """

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  # Step 1:
  if not is_valid_r_tilde(r_tilde, m):
    # Step 1.1:
    return None;

  # Step 2:
  def recursive(x, F):
    l = len(F);

    # Check the timeout.
    timeout.check();

    # Step 2.1:
    if l == 1:
      # Step 2.1.1:
      return {(F[0], x)};

    # Step 2.2:
    F_L = F[:floor(l/2)];
    F_R = F[floor(l/2):];

    # Step 2.3:
    d_L = mpz(prod([q ** (floor(log(c * m) / log(q))) for q in F_R]));
    d_R = mpz(prod([q ** (floor(log(c * m) / log(q))) for q in F_L]));

    x_L = x ** d_L;
    x_R = x ** d_R;

    # Step 2.4:
    return recursive(x_L, F_L).union(recursive(x_R, F_R));

  # Step 3:
  x = g ** r_tilde;
  d = 1;

  # Step 4:
  T = recursive(x, [q for q in prime_range(floor(c * m) + 1)]);

  # Step 5:
  for (q_i, x_i) in T:
    # Step 5.1:
    e_i = 0;
    e_max = floor(log(c * m) / log(q_i));

    # Step 5.2:
    while x_i != 1:
      # Step 5.2.1:
      if e_i == e_max:
        return None;

      # Step 5.2.2:
      x_i = x_i ** q_i;
      d = d * q_i;
      e_i = e_i + 1;

  # Step 6:
  return d * r_tilde;


def algorithm4(
  g : CyclicGroupElement,
  S : set[int | mpz],
  m : int,
  c : int = 1,
  timeout : int | None | Timeout = None) -> set[int | mpz]:

  """ @brief  Returns a subset Sp of S consisting of all r_tilde in S that
              are such that d * r_tilde is a positive integer multiple of r,
              where d is cm-smooth.

      This function implements Alg. 4 from [E24].

      [E24] Ekerå, M.: "On the success probability of quantum order finding".
                       ArXiv 2201.07791v2 (2022).

      As in [E24], d is said to be cm-smooth if d = p1^e1 * .. pk^ek, for
      q1, ..., qk pairwise distinct primes, and e1, ..., ek positive integer
      exponents, if it holds that qi^ei <= cm for all i in [1, k].

      @param g  The element g of order r.

      @param S  A set S of candidates for the integer r_tilde.

      @param m  A positive integer m such that r < 2^m.

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing cm-smooth component d in r = d * r_tilde.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the computation aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @return   A subset Sp of S consisting of all r_tilde in S that are such
                that d * r_tilde is a positive integer multiple of r, where d
                is cm-smooth. """

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  # Step 1:
  e = prod([q ** floor(log(c * m) / log(q))
    for q in prime_range(floor(c * m) + 1)]);
  x = g ** e;

  # Step 2:
  Sp = set();

  # Step 3:
  for tilde_rip in S:

    # Check the timeout.
    timeout.check();

    # Step 3.1:
    if is_valid_r_tilde(tilde_rip, m) and (x ** tilde_rip == 1):
      # Step 3.1.1:
      Sp.add(tilde_rip);

  # Step 4:
  return Sp;


""" @brief  A module for solving the shortest non-zero vector problem (SVP). """

from .enumerate import enumerate;
from .gram_schmidt import gram_schmidt;

from ..vectors import norm2;

from ...utils.timeout import Timeout;

from gmpy2 import mpz;
from gmpy2 import mpq;
from gmpy2 import mpfr;

def solve_svp(
  B : list[list[int | mpz]],
  timeout : int | None | Timeout = None,
  gs : list[list[list[int | mpz | mpq | mpfr]]] | None = None,
  precision : int | None = None) -> list[int | mpz]:

  """ @brief  Let L be the lattice generated by the rows of the integer basis
              matrix B. Then this function computes and returns a shortest
              non-zero vector in L.

      The basis matrix B is represented as a list of lists [b_1, ..., b_n],
      where b_i = [b_i1, ..., b_id] for i = 1, ..., n are the n rows of B.

      It is assumed that the basis B is delta-LLL-reduced.

      It is required that B has integer entries of type int or mpz.

      Furthermore, it is required that B has full rank and is square since the
      enumerate() function that this function calls currently imposes these
      requirements. Said requirements may be relaxed in the future.

      It is assumed that the basis B is reduced, see the lll() function.

      @param B  The matrix B.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the computation aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @param gs   The list [Bs, M], where Bs is the Gram–Schmidt orthogonalized
                  basis for B, and M is the associated matrix of Gram–Schmidt
                  projection factors, as returned by calling

                     [Bs, M] = gram_schmidt(B, precision = precision),

                  or None, in which case this function will make the above call.

                  If you plan to call this functions repeatedly for the same
                  lattice, then time may be saved by not re-computing Bs and M
                  for each call.

                  Note that the basis B is typically LLL-reduced by calling
                  lll() before calling this function. The lll() function can
                  return not only B but also Bs and M (since Bs and M are
                  incrementally computed as a part of the LLL reduction process)
                  allowing you to directly pass them along to this function.

      @param precision  The precision to use when computing the Gram–Schmidt
                        projection factors. May be set to None, in which case
                        the projection factors are represented as exact
                        quotients.

                        Note that this parameter only has an effect if gs is
                        set to None as the Gram–Schmidt orthogonalized basis Bs
                        and the associated matrix M of Gram–Schmidt projection
                        factors are otherwise pre-computed.

      @return   A shortest non-zero vector in L, the lattice generated by the
                rows of the integer basis matrix B. """

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  if None == gs:
    gs = gram_schmidt(B, precision = precision);

  timeout.check();

  # Let v be the first vector in the basis.
  v = B[0];

  # Let R2 be the square norm of v.
  R2 = norm2(v);

  # Enumerate all vectors within distance R of v and see if there is a shorter
  # non-zero vector v' than v. If so, set v to v', update R, and re-start the
  # enumeration (since R is now smaller this is likely more efficient).

  # Set the proceed flag.
  proceed = True;

  while proceed:

    # Unset the proceed flag.
    proceed = False;

    # Check the timeout.
    timeout.check();

    # Start the enumeration.
    for vp in enumerate(B,
                        R2,
                        timeout = timeout,
                        gs = gs):

      tmp = norm2(vp);

      if 0 < tmp < R2:
        v = vp;
        R2 = tmp;

        # Set the proceed flag.
        proceed = True;

        break; # Stop the enumeration.

  return v;


""" @brief  A module for sampling a frequency j from the distribution induced by
            the quantum part of Shor's order-finding algorithm [Shor94] for a
            given order r. This by using the sampling procedure described in
            [E24t] (see in particular Sects. 5.3.5 and 5.4.3).

    This module can also be used to sample from the distribution induced by
    the quantum part of Seifert's variation [Seifert01] of Shor's order-finding
    algorithm for a given order r, depending on how parameters are selected.

    Throughout this module, the algorithms are as described in [E24t], [E24] and
    [E21]. The notation is also inherited from said works.

    [Shor94] Shor, P.W.: "Algorithms for Quantum Computation: Discrete
                          Logarithms and Factoring".
                         In: Proceedings from FOCS '94, pp. 124–134 (1994).

    [Seifert01] Seifert, J.-P.: "Using fewer qubits in Shor's factorization
                                 algorithm via simultaneous Diophantine
                                 approximation". In: CT-RSA 2001.
                                Springer LNCS 2020, pp. 319–227 (2001).

    [E21] Ekerå, M.: "Quantum algorithms for computing general discrete
                      logarithms and orders with tradeoffs".
                     J. Math. Cryptol. 15(1), pp. 359–407 (2021).

    [E24] Ekerå, M.: "On the success probability of quantum order finding".
                     ACM Trans. Quantum Comput. 5(2):11 (2024).

    [E24t] Ekerå, M.: "On factoring integers, and computing discrete logarithms
                       and orders, quantumly".
                      PhD thesis, KTH Royal Institute of Technology (2024). """

import gmpy2;

from gmpy2 import mpz;
from gmpy2 import mpfr;

from gmpy2 import const_pi as mpfr_const_pi;
from gmpy2 import rint as mpfr_round;
from gmpy2 import floor as mpfr_floor;
from gmpy2 import cos as mpfr_cos;
from gmpy2 import gcd;

from ...math.modular import truncmod;

from ...math.random import sample_integer;

from ...utils.timeout import Timeout;

B_DEFAULT_SAMPLE = 10 ** 5;

def optimal_j_for_z_r(
  z : int | mpz,
  r : int | mpz,
  m : int,
  l : int) -> int | mpz:

  """ @brief  Computes and returns the optimal frequency j0(z) for z in [0, r).

      @param z  The peak index z in [0, r).

      @param r  The order r.

      @param m  A positive integer m such that r < 2^m.

      @param l  An integer l on [0, m), such that m + l is the length of the
                control register in the quantum part of the algorithm.

      @return   The optimal frequency j0(z) = round(2^(m + l) / r * z). """

  # Sanity checks.
  if (z < 0) or (z >= r):
    raise Exception("Error: Incorrect parameter z.");

  if (m < 0) or (not (r < (2 ** m))):
    raise Exception("Error: Incorrect parameter m.");

  if (l < 0) or (l > m):
    raise Exception("Error: Incorrect parameter l.");

  # Define precision.
  precision = 2 * (m + l);

  with gmpy2.context(gmpy2.get_context()) as context:

    # Set the precision.
    context.precision = precision;

    # Compute the optimal frequency j0(z).
    z = mpz(z);
    r = mpz(r);

    j0 = mpz(mpfr_round(mpfr(mpz(2 ** (m + l)) * z) / mpfr(r)));

    # Return j0(z).
    return j0;


def sample_j_given_r(
  r : int | mpz,
  m : int,
  l : int,
  B : int = B_DEFAULT_SAMPLE,
  timeout : int | None | Timeout = None,
  verbose : bool = False,
  extended_result : bool = False) -> int | list[int, list[int]] | None:

  """ @brief  Samples a frequency j from the distribution induced by the quantum
              part of Shor's order-finding algorithm [Shor94] for a given order
              r. This by using the sampling procedure described in [E24t] (see
              in particular Sects. 5.3.5 and 5.4.3).

      This fuction can also be used to sample from the distribution induced by
      Seifert's variation [Seifert01] of Shor's order-finding algorithm for a
      given order r, depending on how parameters are selected.

      Throughout this function, the algorithms are as described in [E21], [E24]
      and [E24t]. The notation is also inherited from said works.

      To sample the distribution, one of the r peaks is first picked uniformly
      at random by picking an index z uniformly at random from [0, r). This
      function furthermore selects a pivot uniformly at random from [0, 1).

      For j0(z) the optimal frequency for this peak, it then subtracts the
      probabilities r * P((j0(z) + offset) mod 2^(m + l)) of frequencies that
      are offset by 0, ±1, ±2, ..., ±(B - 1) from the optimal offset from the
      pivot, returning j = (j0(z) + offset) mod 2^(m + l) as soon as pivot <= 0.

      [Shor94] Shor, P.W.: "Algorithms for Quantum Computation: Discrete
                            Logarithms and Factoring".
                           In: Proceedings from FOCS '94, pp. 124–134 (1994).

      [Seifert01] Seifert, J.-P.: "Using fewer qubits in Shor's factorization
                                   algorithm via simultaneous Diophantine
                                   approximation". In: CT-RSA 2001.
                                  Springer LNCS 2020, pp. 319–227 (2001).

      [E21] Ekerå, M.: "Quantum algorithms for computing general discrete
                       logarithms and orders with tradeoffs".
                       J. Math. Cryptol. 15(1), pp. 359–407 (2021).

      [E24] Ekerå, M.: "On the success probability of quantum order finding".
                       ACM Trans. Quantum Comput. 5(2):11 (2024).

      [E24t] Ekerå, M.: "On factoring integers, and computing discrete
                         logarithms and orders, quantumly".
                        PhD thesis, KTH Royal Institute of Technology (2024).

      @param r  The order r.

      @param m  A positive integer m such that r < 2^m.

      @param l  An integer l on [0, m), such that m + l is the length of the
                control register in the quantum part of the algorithm.

      @param B  A parameter B that upper-bounds the offset from the optimal
                frequency j0(z) as explained above.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the sampling procedure aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @param verbose  A flag that may be set to True to print intermediary
                      results when sampling.

      @param extended_result  A flag that may be set to True to not only return
                              the frequency j, but [j, [z, j0(z), offset]].

      @return   The frequency j sampled if the extended_result flag is set to
                False, or [j, [z, j0(z), offset]] if the extended_result flag is
                set to True, or None if sampling failed because the upper bound
                B on the offset from the optimal frequency j0(z) was reached.
      """

  # Sanity checks.
  if (m < 0) or (not (r < (2 ** m))):
    raise Exception("Error: Incorrect parameter m.");

  if (l < 0) or (l > m):
    raise Exception("Error: Incorrect parameter l.");

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  # Define precision.
  precision = 2 * (m + l);
  if precision < 256:
    precision = 256;

  with gmpy2.context(gmpy2.get_context()) as context:

    # Set the precision.
    context.precision = precision;

    # Sample z uniformly at random from [0, r).
    z = mpz(sample_integer(r));
    if verbose:
      print("Sampled z = " + str(z) + "\n");

    r = mpz(r);
    if verbose:
      print("Note: It holds that gcd(z, r) = " + gcd(z, r) + "\n");

    # Compute the optimal frequency j0(z).
    j0 = mpz(mpfr_round(mpfr(mpz(2 ** (m + l)) * z) / mpfr(r)));

    # Explore a region around j0(z).
    pivot = \
      mpfr(mpz(sample_integer(mpz(2 ** precision))), precision) / \
        mpfr(mpz(2 ** precision), precision);

    if verbose:
      print("Sampled pivot =", str(pivot) + "\n");

    def P(j):

      with gmpy2.context(gmpy2.get_context()) as context:

        # Set the precision.
        context.precision = precision;

        L = mpz(mpfr_floor(mpfr(2 ** (m + l)) / mpfr(r)));

        beta = mpz(2 ** (m + l)) % r;

        alpha_r = truncmod(mpz(r * j), mpz(2 ** (m + l)));

        if alpha_r == 0:
          result = mpfr((L ** 2) * r + (2 * L + 1) * beta);
        else:
          theta_r = mpfr(2 * mpfr_const_pi(precision) * alpha_r) / \
            mpfr(2 ** (m + l));

          result  = beta * (1 - mpfr_cos(theta_r * (L + 1)));
          result += mpfr(r - beta) * (1 - mpfr_cos(theta_r * L));
          result /= 1 - mpfr_cos(theta_r);

        result /= mpz(2 ** (2 * (m + l)));

        return result;

    for offset in range(B):

      # Check the timeout.
      timeout.check();

      for sign in [1, -1]:
        if (0 == offset) and (-1 == sign):
          continue;

        j = (j0 + sign * offset) % (2 ** (m + l));
        probability = r * P(j);
        pivot -= probability;

        if verbose:
          print("Offset:", sign * offset, "-- Probability:", probability, \
            "-- Pivot:", pivot, "-- j:", j);

        if pivot <= 0:
          if extended_result:
            return [int(j), [int(z), int(j0), sign * offset]];
          else:
            return int(j);

    return None;



""" @brief  A module for factoring large random RSA integers.

    This module uses Ekerå–Håstad's algorithm to factor RSA integers, as
    described in [EH17], with improvements from [E20] and [E23p].

      [EH17] Ekerå, M. and Håstad, J.: "Quantum Algorithms for Computing Short
                                        Discrete Logarithms and Factoring RSA
                                        Integers.". In: PQCrypto 2017.
                                       Springer LNCS 10346, pp. 347–363 (2017).

      [E20] Ekerå, M.: "On post-processing in the quantum algorithm for
                        computing short discrete logarithms".
                       Des. Codes Cryptogr. 88, pp. 2313–2335 (2020).

      [E23p] Ekerå, M.: "On the success probability of the quantum algorithm for
                         the short DLP". ArXiv 2309.01754v2 (2025). """

from gmpy2 import mpz;

from gmpy2 import ceil;

from ...math.groups import CyclicGroupElement;

def setup_x_given_g_N(
    g : CyclicGroupElement,
    N : int | mpz) -> CyclicGroupElement:

  """ @brief  Sets up x = g^d' for d' = (N - 1) / 2 - 2^(l - 1) given g and N,
              for N the product of two large random distinct l-bit primes.

      This is a convenience function for Ekerå–Håstad's algorithm [EH17] that
      factors a large random RSA integer N = pq into p and q.

      More specifically, as is explained in App. A.2 of [E20], it holds that

        x = g^d' = g^((N - 1) / 2 - 2^(l - 1))
                 = g^((p - 1) / 2 + (q - 1) / 2 - 2^(l - 1)) = g^d.

      Provided that the order r of g is sufficiently large, it holds that

        d = d' mod r = (p - 1) / 2 + (q - 1) / 2 - 2^(l - 1),

      allowing d to be computed as the discrete logarithm of x to the base g.

      This may be done efficiently by using Ekerå–Håstad's quantum algorithm
      that computes short discrete logarithms in groups of unknown order.

      Given d = (p - 1) / 2 + (q - 1) / 2 - 2^(l - 1) and N = pq, it is then
      trivial to compute p and q by solving a quadratic equation.

      This convenience function computes x in the above procedure given g and N.

      [EH17] Ekerå, M. and Håstad, J.: "Quantum Algorithms for Computing Short
                                        Discrete Logarithms and Factoring RSA
                                        Integers.". In: PQCrypto 2017.
                                       Springer LNCS 10346, pp. 347–363 (2017).

      [E20] Ekerå, M.: "On post-processing in the quantum algorithm for
                        computing short discrete logarithms".
                       Des. Codes Cryptogr. 88, pp. 2313–2335 (2020).

      @param g  The generator g selected uniformly at random from the
                multiplicative group of the ring of integers modulo N.

      @param N  The integer N = pq, for p and q two large distinct random
                l-bit prime numbers.

      @return   The element x = g^((N - 1) / 2 - 2^(l - 1)). """

  # Convert inputs.
  N = mpz(N);

  # Sanity checks.
  if 1 != (N % 2):
    raise Exception("Error: Expected N to be odd.");

  # Compute the bit length l of p and q where N = pq: We have that
  #
  #   p, q >= 2^(l-1) => N = pq = 2^(2(l-1))
  #                        = 2^(2l-2),
  #
  # i.e. N is an n >= 2l - 1 bit integer. At the same time, we have that
  #
  #   p, q < 2^l => N = pq < 2^2l,
  #
  # i.e. N is an n <= 2l bit integer. Hence, for n the bit length of N, it must
  # be that l = ceil(n / 2), so we may compute l from n in this manner:

  l = int(ceil(N.bit_length() / 2));

  # Select the exponent d' as explained in [E20] and compute x.
  x = g ** ((N - 1) // 2 - 2 ** (l - 1));

  # Return x.
  return x;


def setup_d_given_p_q(
  p : int | mpz,
  q : int | mpz) -> int:

  """ @brief  Sets up d = (p - 1) / 2 + (q - 1) / 2 - 2^(l - 1) given p and q.

      This is a convenience function for Ekerå–Håstad's algorithm [EH17] that
      factors a large random RSA integer N = pq into p and q.

      More specifically, as is explained in App. A.2 of [E20], it holds that

        x = g^d' = g^((N - 1) / 2 - 2^(l - 1))
                 = g^((p - 1) / 2 + (q - 1) / 2 - 2^(l - 1)) = g^d

      Provided that the order r of g is sufficiently large, it holds that

        d = d' mod r = (p - 1) / 2 + (q - 1) / 2 - 2^(l - 1),

      allowing d to be computed as the discrete logarithm of x to the base g.

      This may be done efficiently by using Ekerå–Håstad's quantum algorithm
      that computes short discrete logarithms in groups of unknown order.

      Given d = (p - 1) / 2 + (q - 1) / 2 - 2^(l - 1) and N = pq, it is then
      trivial to compute p and q by solving a quadratic equation.

      This convenience function computes d in the above procedure given p and q.

      [EH17] Ekerå, M. and Håstad, J.: "Quantum Algorithms for Computing Short
                                        Discrete Logarithms and Factoring RSA
                                        Integers.". In: PQCrypto 2017.
                                       Springer LNCS 10346, pp. 347–363 (2017).

      [E20] Ekerå, M.: "On post-processing in the quantum algorithm for
                        computing short discrete logarithms".
                       Des. Codes Cryptogr. 88, pp. 2313–2335 (2020).

      @param p  A large random l-bit prime.
      @param q  A large random l-bit prime not equal to p.

      @return   The logarithm d = ((p - 1) / 2) + ((q - 1) / 2) - 2^(l - 1). """

  # Sanity checks.
  if p == q:
    raise Exception("Error: The primes p and q must be distinct.");

  if ((p % 2) != 1) or ((q % 2) != 1):
    raise Exception("Error: The primes p and q must be odd.");

  # Compute the bit length l of p and q.
  l = p.bit_length();
  if q.bit_length() != l:
    raise Exception("Error: The primes p and q must be of equal bit length.");

  # Compute d as explained in [E20].
  d = ((p - 1) // 2) + ((q - 1) // 2) - 2 ** (l - 1);

  # Return d;
  return int(d);


""" @brief  A module for finding the largest power of two that divides an
            integer. """

from gmpy2 import mpz;

def kappa(x : int | mpz) -> int:

  """ @brief  Given an integer x, this function returns t such that x = 2^t * o,
              for o odd.

      @param x  The integer x.

      @return   A non-negative integer t such that x = 2^t * o, for o odd. """

  if x == 0:
    return 0;

  t = 0;

  while (x % 2) == 0:
    t += 1;
    x //= 2;

  return t;


""" @brief  A module for computing prime ranges and prime power products, for
            sampling random primes and for testing if integers are smooth. """

from gmpy2 import mpz;

from gmpy2 import is_prime; # Imported also to be available to other modules.

from math import ceil;

from .random import sample_l_bit_integer;

def prime_range(B : int) -> list[int]:

  """ @brief  Returns an ordered list of all primes less than B.

      @param B  The upper bound B.

      @return   An ordered list of all primes less than B. """

  sieve = [True for _ in range(0, B)];
  primes = [];

  p = 2;
  while p < B:
    primes.append(p);

    for n in range(2, ceil(B / p)):
      sieve[n * p] = False;

    p += 1;
    while (p < B) and (not sieve[p]):
      p += 1;

  return primes;


def prime_power_product(B : int) -> int | mpz:

  """ @brief  Returns the product of q^e, as q runs over all primes <= B, for e
              the largest non-negative integer exponent such that q^e <= B.

      @param B  The upper bound B.

      @return   The product of q^e, as q runs over all primes <= B, for e the
                largest non-negative integer exponent such that q^e <= B."""

  factor = 1;

  for q in prime_range(B + 1):
    q_pow_e = q;

    while True:
      tmp = q_pow_e * q;
      if tmp > B:
        break;
      q_pow_e = tmp;

    factor *= q_pow_e;

  return factor;


def is_B_smooth(d : int | mpz, B : int) -> bool:

  """ @brief  Tests if the integer d is B-smooth.

      As in [E24], d is said to be B-smooth if d = p1^e1 * .. pk^ek, for
      q1, ..., qk pairwise distinct primes, and e1, ..., ek positive integer
      exponents, if it holds that qi^ei <= B for all i in [1, k].

      [E24] Ekerå, M.: "On the success probability of quantum order finding".
                       ACM Trans. Quantum Comput. 5(2):11 (2024).

      @param d  The integer d.

      @param B  The upper bound B.

      @return   True if d is B-smooth, False otherwise. """

  if d <= 0:
    raise Exception("Error: Incorrect parameters.");

  for q in prime_range(B + 1):
    q_pow_e = 1;

    while (d % q) == 0:
      q_pow_e *= q;
      if q_pow_e > B:
        return False;
      d //= q;

  return d == 1;


def sample_l_bit_prime(l : int) -> int | mpz:

  """ @brief  Returns an l-bit prime selected uniformly at random from the set
              of all such primes.

      @remark   This function calls sample_l_bit_integer() to select l-bit
                integers uniformly at random from the set of all l-bit integers.
                In practice, sample_l_bit_integer() returns an l-bit integer
                that may be conjectured to be indistinguishable from an integer
                that is selected uniformly at random from said set.

      @remark   This function assumes that the is_prime() function performs a
                deterministic primality test. In practice, this test is likely
                probabilistic, but the probability of incorrectly classifying
                an composite as a prime is so small that the test may be
                conjectured to be indistinguishable from a determinstic test.

      @param l  The bit length l of the prime to sample.

      @return   An l-bit prime selected uniformly at random from the set of all
                such primes. """

  while True:
    p = mpz(sample_l_bit_integer(l));
    if p.is_prime():
      return p;


""" @brief  A module for sampling frequency pairs from the distribution induced
            by the quantum part of Ekerå–Håstad's algorithm [EH17] for finding
            short discrete logarithms in groups of unknown order.

    [EH17] Ekerå, M. and Håstad, J.: "Quantum Algorithms for Computing Short
                                      Discrete Logarithms and Factoring RSA
                                      Integers.". In: PQCrypto 2017.
                                     Springer LNCS 10346, pp. 347–363 (2017).
"""

import gmpy2;

from gmpy2 import mpz;
from gmpy2 import mpq;
from gmpy2 import mpfr;

from gmpy2 import const_pi as mpfr_const_pi;
from gmpy2 import cos as mpfr_cos;

from ...math.modular import truncmod;

from ...math.random import sample_integer;

from ...utils.timeout import Timeout;

B_DEFAULT = 10 ** 5;

def sample_j_k_given_d_r_tau(
  d : int | mpz,
  r : int | mpz | None,
  m : int,
  l : int,
  tau : int,
  timeout : int | None | Timeout = None,
  verbose : bool = False,
  extended_result : bool = False) -> list[int | mpz] | None:

  """ @brief  Samples a frequency pair (j, k) from the distribution induced by
              the quantum part of Ekerå–Håstad's algorithm [EH17] for finding a
              short discrete logarithm d in a group of unknown order r.

      The sampling procedure is described in Sect. 5.6.5 of [E24t].

      To sample the distribution, j is first picked uniformly at random from
      [0, 2^(m + l)). A pivot is then selected uniformly at random from [0, 1),
      and the optimal frequency k0(j) for k computed.

      For offsets 0, ±1, ±2, ..., ±(B - 1) from k0(j) the probabilities
      2^(m+l) * P(j, k = (k0(j) + offset) mod 2^l) of observing (j, k) are then
      subtracted from the pivot, and (j, k) returned as soon as pivot <= 0
      provided that (j, k) is tau-good by Def. 1 in [E23p]. Otherwise, None
      is returned. The bound B is setup as a function of tau so that all
      tau-good pairs (j, k) are included in the search.

      Note that it follows from the analysis in [E20] and [E23p] that j is
      distributed uniformly at random when r >= 2^(m + l) + (2^l - 1) * d. This
      function checks that this requirement is respected provided that r is
      included in the function call.

      [EH17] Ekerå, M. and Håstad, J.: "Quantum Algorithms for Computing Short
                                        Discrete Logarithms and Factoring RSA
                                        Integers.". In: PQCrypto 2017.
                                       Springer LNCS 10346, pp. 347–363 (2017).

      [E20] Ekerå, M.: "On post-processing in the quantum algorithm for
                        computing short discrete logarithms".
                       Des. Codes Cryptogr. 88, pp. 2313–2335 (2020).

      [E23p] Ekerå, M.: "On the success probability of the quantum algorithm for
                         the short DLP". ArXiv 2309.01754v2 (2025).

      [E24t] Ekerå, M.: "On factoring integers, and computing discrete
                         logarithms and orders, quantumly".
                        PhD thesis, KTH Royal Institute of Technology (2024).

      @param d  The discrete logarithm d in [1, r).

      @param r  The order r. Used to check that r >= 2^(m + l) + (2^l - 1) * d.
                May be set to None in which case the check is not performed.

      @param m  A positive integer m such that d < 2^m.

      @param l  A positive integer l such that the control registers in the
                quantum part of the algorithm are of lengths m + l and l qubits,
                respectively.

                It is required that r >= 2^(m + l) + (2^l - 1) * d as this
                function is based on the analysis in [E20] that imposes
                this requirement so as to simplify the analysis.

      @param tau  The parameter tau. An integer on (0, l].

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the sampling procedure aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @param verbose  A flag that may be set to True to print intermediary
                      results when sampling.

      @param extended_result  A flag that may be set to True to not only return
                              the frequency pair (j, k) as a list [j, k], but
                              [[j, k], [k0(j), offset]].

      @return   The frequency pair [j, k] sampled if the extended_result flag is
                set to False, or [[j, k], [k0(j), offset]] if the
                extended_result flag is set to True, or None if sampling failed
                because the upper bound B on the offset from the optimal
                frequency k0(j) was reached, or because (j, k) is not tau-good.
  """

  result = sample_j_k_given_d_r(d,
                                r,
                                m,
                                l,
                                B = (2 ** tau) + 2,
                                timeout = timeout,
                                verbose = verbose,
                                extended_result = True);

  if None == result:
    return None;

  [[j, k], _] = result;

  alpha = truncmod(mpz(d * j + (2 ** m) * k), mpz(2 ** (m + l)));
  if not (abs(alpha) < mpz(2 ** (m + tau))):
    return None;

  if extended_result:
    return result;
  else:
    return [j, k];


def sample_j_k_given_d_r(
  d : int | mpz,
  r : int | mpz | None,
  m : int,
  l : int,
  B : int = B_DEFAULT,
  timeout : int | None | Timeout = None,
  verbose : bool = False,
  extended_result : bool = False) -> list[int | mpz] | None:

  """ @brief  Samples a frequency pair (j, k) from the distribution induced by
              the quantum part of Ekerå–Håstad's algorithm [EH17] for finding a
              short discrete logarithm d in a group of unknown order r.

      The sampling procedure is described in Sect. 5.6.5 of [E24t].

      To sample the distribution, j is first picked uniformly at random from
      [0, 2^(m + l)). A pivot is then selected uniformly at random from [0, 1),
      and the optimal frequency k0(j) for k computed.

      For offsets 0, ±1, ±2, ..., ±(B - 1) from k0(j) the probabilities
      2^(m+l) * P(j, k = (k0(j) + offset) mod 2^l) of observing (j, k) are then
      subtracted from the pivot, and (j, k) returned as soon as pivot <= 0.

      Note that it follows from the analysis in [E20] and [E23p] that j is
      distributed uniformly at random when r >= 2^(m + l) + (2^l - 1) * d. This
      function checks that this requirement is respected provided that r is
      included in the function call.

      [EH17] Ekerå, M. and Håstad, J.: "Quantum Algorithms for Computing Short
                                        Discrete Logarithms and Factoring RSA
                                        Integers.". In: PQCrypto 2017.
                                       Springer LNCS 10346, pp. 347–363 (2017).

      [E20] Ekerå, M.: "On post-processing in the quantum algorithm for
                        computing short discrete logarithms".
                       Des. Codes Cryptogr. 88, pp. 2313–2335 (2020).

      [E23p] Ekerå, M.: "On the success probability of the quantum algorithm for
                         the short DLP". ArXiv 2309.01754v2 (2025).

      [E24t] Ekerå, M.: "On factoring integers, and computing discrete
                         logarithms and orders, quantumly".
                        PhD thesis, KTH Royal Institute of Technology (2024).

      @param d  The discrete logarithm d in [1, r).

      @param r  The order r. Used to check that r >= 2^(m + l) + (2^l - 1) * d.
                May be set to None in which case the check is not performed.

      @param m  A positive integer m such that d < 2^m.

      @param l  A positive integer l such that the control registers in the
                quantum part of the algorithm are of lengths m + l and l qubits,
                respectively.

                It is required that r >= 2^(m + l) + (2^l - 1) * d as this
                function is based on the analysis in [E20] that imposes
                this requirement so as to simplify the analysis.

      @param B  A parameter B that upper-bounds the offset from the optimal
                frequency k0(j) as explained above.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the sampling procedure aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @param verbose  A flag that may be set to True to print intermediary
                      results when sampling.

      @param extended_result  A flag that may be set to True to not only return
                              the frequency pair (j, k) as a list [j, k], but
                              [[j, k], [k0(j), offset]].

      @return   The frequency pair [j, k] sampled if the extended_result flag is
                set to False, or [[j, k], [k0(j), offset]] if the
                extended_result flag is set to True, or None if sampling failed
                because the upper bound B on the offset from the optimal
                frequency k0(j) was reached. """

  # Sanity checks.
  if (m <= 0) or (not (d < (2 ** m))):
    raise Exception("Error: Incorrect parameter m.");

  if l <= 0:
    raise Exception("Error: Incorrect parameter l.");

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  # Check the requirement that r >= 2 ** (m + l) + (2 ** l - 1) * d.
  if (r != None) and (r < 2 ** (m + l) + (2 ** l - 1) * d):
    raise Exception("Error: It is required that r >= 2^(m + l) + (2^l - 1) * d "
      + "as this function is based on the analysis in [E20] that imposes this "
      + "requirement so as to simplify the analysis.");

  # Sample the frequency j uniformly at random from [0, 2^(m + l)).
  j = mpz(sample_integer(2 ** (m + l)));
  if verbose:
    print("Sampled j =", str(j));

  # Compute the optimal frequency k0(j).
  k0 = mpz(-round(mpq(d * j, 2 ** m))) % mpz(2 ** l);
  if verbose:
    print("Computed k0(j) =", k0);

  # Define the precision.
  precision = 2 * (m + 2 * l);
  if precision < 256:
    precision = 256;

  with gmpy2.context(gmpy2.get_context()) as context:

    # Set the precision.
    context.precision = precision;

    # Explore a region around k0(j).
    pivot = \
      mpfr(mpz(sample_integer(mpz(2 ** precision))), precision) / \
        mpfr(mpz(2 ** precision), precision);

    if verbose:
      print("Sampled pivot =", str(pivot) + "\n");

    def P(j, k):

      with gmpy2.context(gmpy2.get_context()) as context:

        # Set the precision.
        context.precision = precision;

        # Compute the probability as described in Sect. 3.2 of [E20].
        alpha = truncmod(mpz(d * j + (2 ** m) * k), mpz(2 ** (m + l)));

        if alpha == 0:
          # Use the expression for P(0), see Sect. 3.2 in [E20].
          result  = ((2 ** (m + l)) - ((2 ** l) - 1) * d) * (2 ** (2 * l));
          result += (((2 ** l) * d) / 3) * ((2 ** l) - 1) * ((2 ** (l + 1)) - 1);
          result /= 2 ** (2 * (m + 2 * l));

        else:
          # Use the expression for P(theta), see Sect. 3.2 in [E20].
          theta = mpfr(2 * mpfr_const_pi(precision) * alpha) / mpz(2 ** (m + l));

          result  = mpfr_cos(mpz((2 ** l) - 1) * theta) - \
              mpfr_cos(mpz(2 ** l) * theta);
          result /= 1 - mpfr_cos(theta);
          result -= 1;
          result /= 2;
          result  = mpz((2 ** l) - 1) - result;
          result *= mpz(2 * d);
          result += (mpz(2 ** (m + l)) - mpz((2 ** l) - 1) * d) * \
            (1 - mpfr_cos(mpz(2 ** l) * theta));
          result /= 1 - mpfr_cos(theta);
          result /= mpz(2 ** (2 * (m + 2 * l)));

        return result;

    # Sample k.
    if verbose:
      print("Sampling k...");

    for offset in range(B):

      # Check the timeout.
      timeout.check();

      for sign in [1, -1]:

        if (0 == offset) and (-1 == sign):
          continue;

        k = mpz(k0 + sign * offset) % mpz(2 ** l);
        probability = mpz(2 ** (m + l)) * P(j, k);
        pivot -= probability;

        if verbose:
          print("Offset:", sign * offset, "-- Probability:", probability, \
                  "-- Pivot:", pivot, "-- k:", k);

        if pivot <= 0:
          if extended_result:
            return [[int(j), int(k)], [int(k0), sign * offset]];
          else:
            return [int(j), int(k)];

    return None;



""" @brief  A module for matrices. """

from typing import Any;

from gmpy2 import mpz;
from gmpy2 import mpq;

from copy import deepcopy;


def dimensions(B : list[list[Any]]) -> tuple[int, int]:

  """ @brief  Returns the dimensions of the matrix B.

      @param B  The matrix B.

      @return   The dimensions (n, d) of the n x d-dimensional matrix B. """

  n = len(B);
  d = len(B[0]);

  return (n, d);


def invert(B : list[list[int | mpz | mpq]]) -> list[list[int | mpz | mpq]]:

  """ @brief  Inverts a full-rank rational n x n-matrix B.

      @param B  The matrix B = [b_1, ..., b_n], for b_i = [b_i1, ..., b_in] for
                i = 1, ..., n the n rows of B. The entries b_ij must be of type
                int or mpz for integer B, or of type mpq for rational B.

      @return   The inverse of the matrix B. """

  (n, d) = dimensions(B);

  if n != d:
    raise ValueError("Error: The matrix B must be a square matrix.");

  # Duplicate B.
  B = deepcopy(B);

  # Set B^-1 to I.
  B_inv = [[0 for _ in range(d)] for _ in range(n)];
  for i in range(n):
    B_inv[i][i] = 1;

  # The idea is now to apply row manipulations on B to transform it to I, and to
  # apply the exact same row manipulations to B_inv to transform it to B^-1.

  for column in range(n):

    # Find the first unused row in B that is non-zero in the column.
    row = column;
    while row < n:
      if B[row][column] != 0:
        break;
      row += 1;

    if row == n:
      raise ValueError("Error: The matrix B is not full rank.");

    # Force the entry in said row and column to one.
    c = mpq(1, B[row][column]);
    for j in range(d):
      B[row][j] *= c;
      B_inv[row][j] *= c;

    # Force the entries in all other rows in said column to zero.
    for i in range(n):
      if row == i:
        continue;

      c = B[i][column];
      for j in range(d):
        B[i][j] -= c * B[row][j];
        B_inv[i][j] -= c * B_inv[row][j];

    # Move the row to the right position if necessary.
    if row != column:
      tmp = B[row];
      B[row] = B[column];
      B[column] = tmp;

      tmp = B_inv[row];
      B_inv[row] = B_inv[column];
      B_inv[column] = tmp;

  # Return the inverse.
  return B_inv;


def solve_left(
  B : list[list[int | mpz | mpq]],
  t : list[int | mpz | mpq],
  B_inv : list[list[int | mpz | mpq]] | None = None) -> \
    list[int | mpz | mpq]:

  """ @brief  Given a full-rank rational n x n matrix B, and an n-dimensional
              rational row vector t such that c B = t where c is also a rational
              n-dimensional row vector, this function returns c.

      @param B  The matrix B = [b_1, ..., b_n], for b_i = [b_i1, ..., b_in] for
                i = 1, ..., n the n rows of B. The entries b_ij for i, j in
                1, ..., n must be of type int, mpz or mpq.

      @param t  The n-dimensional rational row vector t = [t_1, ..., t_n]. The
                entries t_i for i = 1, ..., n must be of type int, mpz or mpq.

      @param B_inv  The inverse of the matrix B as computed by calling the
                    function inverse(B), or None in which case this function
                    will call said function to compute the inverse.

      @return   The n-dimensional row vector c = [c_1, ..., c_n] such that
                c B = t. """

  (n, d) = dimensions(B);

  # Check that B is a square matrix.
  if n != d:
    raise Exception("Error: The matrix B must be a square matrix.");

  # We seek to solve c B = t for t. We do so simply by multiplying to the right
  # with the inverse B^-1 of B, yielding c = c B B^-1 = t B^-1.
  if None == B_inv:
    # Compute the inverse B^-1 of B.
    B_inv = invert(B);

  # Multiply t to the right by the inverse to form c = t B^-1.
  c = [sum([t[i] * B_inv[i][j] for i in range(n)]) for j in range(d)]

  # Return c.
  return c;


""" @brief  A module for continued fractions. """

from gmpy2 import mpz;
from gmpy2 import mpq;
from gmpy2 import mpfr;

def continued_fractions(
    j : int | mpz,
    m : int,
    l : int,
    denominator_bound : bool = None) -> list[int | mpz]:

  """ @brief  Expands j / 2^(m + l) in continued fractions, and returns an
              ordered list of all denominators < 2^((m + l) / 2), unless an
              upper bound on the denominator is explicitly specified.

      @param j  The frequency j. An integer on [0, 2^(m+l)).

      @param m  A positive integer.

      @param l  A non-negative integer.

      @param denominator_bound  An upper bound on the denominator. If set to
                                None, as is the default, the bound is taken to
                                be 2^((m + l) / 2).

      @return   An ordered list of all denominators < 2^((m + l) / 2), unless an
                upper bound on the denominator is explicitly specified. """

  denominators = [];

  if None == denominator_bound:
    denominator_bound = mpfr(2) ** mpfr((m + l) / 2);

  fraction = mpq(mpz(j), mpz(2 ** (m + l)));

  km1 = mpz(0);
  km2 = mpz(1);

  while True:
    integer_part = fraction.numerator // fraction.denominator; # = floor(f)

    denominator = integer_part * km1 + km2;
    if denominator >= denominator_bound:
      break;

    denominators.append(denominator);

    # Update the recursion.
    km2 = km1;
    km1 = denominator;

    # Update the fraction: The next fraction is 1 / (f - floor(f)).
    fraction = fraction - integer_part;
    if fraction == 0:
      # The best approximation has been reached.
      break;

    fraction = mpz(1) / fraction;

  return denominators;


""" @brief  A module for solving frequency pairs yielded by the quantum part of
            Ekerå–Håstad's algorithm [EH17] for computing short discrete
            logarithms. This by using the post-processing algorithms in [E20],
            [E23p] and [E24t] (see Sect. 5.6).

    [E20] Ekerå, M.: "On post-processing in the quantum algorithm for computing
                      short discrete logarithms".
                     Des. Codes Cryptogr. 88, pp. 2313–2335 (2020).

    [E23p] Ekerå, M.: "On the success probability of the quantum algorithm for
                       the short DLP". ArXiv 2309.01754v2 (2025).

    [E24t] Ekerå, M.: "On factoring integers, and computing discrete
                       logarithms and orders, quantumly".
                      PhD thesis, KTH Royal Institute of Technology (2024). """

from enum import Enum;

from math import ceil;

from gmpy2 import mpz;
from gmpy2 import mpfr;
from gmpy2 import mpq;

from gmpy2 import floor as mpfr_floor;
from gmpy2 import rint as mpfr_round;
from gmpy2 import ceil as mpfr_ceil;

from gmpy2 import sqrt as mpfr_sqrt;

from ....math.lattices.babai import babai;

from ....math.lattices.enumerate import enumerate as enumerate_lattice;

from ....math.lattices.lagrange import lagrange;

from ....math.lattices.cvp import solve_cvp;

from ....math.lattices.lll import lll;
from ....math.lattices.lll import LLL_DEFAULT_DELTA;
from ....math.lattices.lll import LLL_DEFAULT_REDUCED_PRECISION;

from ....math.matrices import solve_left;

from ....utils.timeout import Timeout;

from ....math.groups import CyclicGroupElement;

from ....math.modular import truncmod;


class EnumerationOptions(Enum):

  """ @brief  An enumeration of options for solving a list of n frequency pairs
              [[j_1, k_1], ..., [j_n, k_n]] for d.

      The meaning of the options are explained in the documentation for the
      solve_multiple_j_k_for_d() function. """

  TRUE = True;

  FALSE = False;

  CVP = "CVP";

  BOUNDED_BY_TAU = "BOUNDED_BY_TAU";


def expected_u_for_j_k_d(
  j : int | mpz,
  k : int | mpz,
  m : int,
  l : int,
  d : int | mpz,
  tau : int = 0) -> list[int | mpz]:

  """ @brief  Computes the vector u that is associated with a given frequency
              pair (j, k) and logarithm d, when using the post-processing
              algorithm from [E23p] to post-process the output of the quantum
              part of Ekerå–Håstad's algorithm [EH17].

      [EH17] Ekerå, M. and Håstad, J.: "Quantum Algorithms for Computing Short
                                        Discrete Logarithms and Factoring RSA
                                        Integers.". In: PQCrypto 2017.
                                       Springer LNCS 10346, pp. 347–363 (2017).

      [E23p] Ekerå, M.: "On the success probability of the quantum algorithm for
                         the short DLP". ArXiv 2309.01754v2 (2025).

      Recall that the vector v = (truncmod(-2^m k, 2^(m+l)), 0), and that the
      vector u, which is in the lattice L^tau, is such that the difference
      u - v = (truncmod(dj - 2^m k, 2^(m+l)), 2^tau d).

      Given j, k, m, l, d and tau this function returns the vector u.

      @param j  The frequency j. An integer on [0, 2^(m + l)).

      @param k  The frequency k. An integer on [0, 2^l).

      @param m  A positive integer m such that d < 2^m.

      @param l  A positive integer l. The control registers in the quantum part
                of the algorithm are of lengths m + l and l qubits,
                respectively.

      @param d  The discrete logarithm d.

      @param tau  An integer tau on [0, l]. Used to scale the basis for the
                  lattice L^tau that is used in the post-processing, and that is
                  generated by the vectors (j, 2^tau) and (2^(m + l), 0).

      @return   The vector u. """

  return expected_u_for_multiple_j_k_d([[j, k]], m, l, d, tau)


def expected_u_for_multiple_j_k_d(
  j_k_list : list[list[int | mpz]],
  m : int,
  l : int,
  d : int | mpz,
  tau : int = 0) -> list[int | mpz]:

  """ @brief  Computes the vector u that is associated with a given list of n
              frequency pairs [[j_1, k_1], ..., [j_n, k_n]] and logarithm d,
              when using the post-processing algorithm from [E20] or [E23p]
              to post-process the output of the quantum part of Ekerå–Håstad's
              algorithm [EH17].

      [EH17] Ekerå, M. and Håstad, J.: "Quantum Algorithms for Computing Short
                                        Discrete Logarithms and Factoring RSA
                                        Integers.". In: PQCrypto 2017.
                                       Springer LNCS 10346, pp. 347–363 (2017).

      [E20] Ekerå, M.: "On post-processing in the quantum algorithm for
                        computing short discrete logarithms".
                       Des. Codes Cryptogr. 88, pp. 2313–2335 (2020).

      [E23p] Ekerå, M.: "On the success probability of the quantum algorithm for
                         the short DLP". ArXiv 2309.01754v2 (2025).

      Recall that the vector

        v = (truncmod(-2^m k_1, 2^(m+l)), ..., truncmod(-2^m k_n, 2^(m+l)), 0),

      and that the vector u, which is in the lattice L^tau, is such that the
      difference

        u - v = (truncmod(dj_1 - 2^m k_1, 2^(m+l)),
                 ...,
                 truncmod(dj_n - 2^m k_n, 2^(m+l)), 2^tau d).

      Given [[j_1, k_1], ..., [j_n, k_n]], m, l, d and tau this function returns
      the vector u.

      @param j_k_list  The list of n frequencies [[j_1, k_1], ..., [j_n, k_n]].

      @param m  A positive integer m such that d < 2^m.

      @param l  A positive integer l. The control registers in the quantum part
                of the algorithm are of lengths m + l and l qubits,
                respectively.

      @param d  The discrete logarithm d.

      @param tau  An integer tau on [0, l]. Used to scale the basis for the
                  lattice L^tau that is used in the post-processing, and that is
                  generated by the vector (j_1, ..., j_n, 2^tau), and by the n
                  vectors (2^(m + l), 0, ..., 0) thru (0, ..., 0, 2^(m + l), 0).

      @return   The vector u. """

  # Initial setup.
  n = len(j_k_list);

  # Compute the vector u component by component.
  u = [];

  for i in range(n):
    [j_i, k_i] = j_k_list[i];

    v_i = truncmod(-(mpz(2 ** m) * k_i), mpz(2 ** (m + l)));

    t1 = mpz(d * j_i - v_i);
    t2 = truncmod(t1, mpz(2 ** (m + l)));
    mp = mpz(round(mpq(t1 - t2, mpz(2 ** (m + l)))));

    u.append(d * j_i - mpz(2 ** (m + l)) * mp);

  u.append(mpz(2 ** tau) * mpz(d));

  # Return u.
  return u;


def solve_j_k_for_d(
  j : int | mpz,
  k : int | mpz,
  m : int,
  l : int,
  g : CyclicGroupElement,
  x : CyclicGroupElement,
  tau : int,
  t : int = None,
  c : int = 1,
  timeout : int | None | Timeout = None,
  verbose : bool = False) -> int | None:

  """ @brief  Attempts to compute the short discrete logarithm d = log_g x given
              a frequency pair (j, k) yielded by the quantum part of
              Ekerå–Håstad's algorithm [EH17] by using the post-processing
              algorithms in [E23p].

      [EH17] Ekerå, M. and Håstad, J.: "Quantum Algorithms for Computing Short
                                        Discrete Logarithms and Factoring RSA
                                        Integers.". In: PQCrypto 2017.
                                       Springer LNCS 10346, pp. 347–363 (2017).

      [E23p] Ekerå, M.: "On the success probability of the quantum algorithm for
                         the short DLP". ArXiv 2309.01754v2 (2025).

      This function implements the enumeration procedure in Alg. 1 in [E23p]. It
      is guaranteed to recover d if (j, k) is a tau-good pair, and if j is such
      that the lattice L^tau is t-balanced, see Defs. 1–3 in [E23p].

      As shown in Thm. 1 in [E23p], the probability that (j, k) fulfills these
      conditions for t and tau is at least

          max(0, 1 - f(2^tau)) * max(0, 1 - 2^(Delta - 2(t-1) - tau))

      for f(B) = 1 / (B - 1) - 1 / (2 (B - 1)^2) - 1 / (6 (B - 1)^3), and for
      Delta = m - l on [0, m), for m, l parameters to the quantum algorithm.

      Furthermore, as shown in Thm. 1 in [E23p], at most 2^3 c sqrt(N) group
      operations must be performed during the enumeration provided that a few
      group elements are first pre-computed, and provided that there is space to
      store at most 2^3 sqrt(N) / c integers in a lookup table, for c a positive
      integer constant, and for N = 2^(Delta + tau + 1) + 2^(tau + t + 2) + 2.

      @param j  The frequency j. An integer on [0, 2^(m + l)).

      @param k  The frequency k. An integer on [0, 2^l).

      @param m  A positive integer m such that d < 2^m.

      @param l  An integer l = m - Delta for Delta an integer on [0, m). The two
                control registers in the quantum part of the algorithm are of
                lengths m + l and l qubits, respectively.

      @param g  The group element g.

      @param x  The group element x = g^d.

      @param tau  An integer tau on (0, l]. Used to scale the basis for the
                  lattice L^tau that is used in the post-processing, and that
                  is generated by the vectors (j, 2^tau) and (2^(m + l), 0).

      @param t  The parameter t. An integer on [0, m). May be set to None, in
                which case t will be implicitly selected so that the lattice
                L^tau is t-balanced. If t is not set to None, this function
                will return None if the lattice L^tau is not t-balanced.

      @param c  The constant c. A positive integer.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the computation aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @param verbose  A flag that may be set to True to print intermediary
                      results and status updates when executing the
                      post-processing algorithm.

      @return   The logarithm d, or None, if solving for d fails. """

  # Sanity checks.
  if not (0 < l <= m):
    raise Exception("Error: Incorrect parameter:",
                      "l must be an integer on (0, m].");

  if not (0 <= tau <= l):
    raise Exception("Error: Incorrect parameter:",
                      "tau must be an integer on [0, l].");

  if not (0 <= t < m):
    raise Exception("Error: Incorrect parameter:",
                      "t must be an integer on [0, m).");

  if not (c >= 1):
    raise Exception("Error: Incorrect parameter:",
                      "c must be a positive integer.");

  if not (0 <= j < (2 ** (m + l))):
    raise Exception("Error: Incorrect parameter:",
                      "j must be an integer on [0, 2^(m + l)).");

  if not (0 <= k < (2 ** l)):
    raise Exception("Error: Incorrect parameter:",
                      "k must be an integer on [0, 2^l).");

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  j = mpz(j);
  k = mpz(k);

  # Step 1:
  def meet_in_the_middle(
      g : CyclicGroupElement,
      x : CyclicGroupElement,
      nu1 : int | mpz,
      nu2 : int | mpz,
      B1 : int | mpz,
      B2 : int | mpz,
      s1 : mpz,
      s2 : mpz,
      mu : mpq,
      c : int = 1) -> int | None:

    # Step 1.2:
    g1 = g ** s1;
    g2 = g ** s2;
    w = (g1 ** nu1) * (g2 ** nu2) * (x ** -1);

    # Step 1.2:
    n = c * mpz(mpfr_round(mpfr_sqrt(B1 / (B2 + 1))));

    # Step 1.3:
    if verbose:
      print("The first stage begins:", \
              2 * (int(ceil(B1 / n)) - 1), "operation(s)");

    T = dict();
    T[g ** 0] = 0;

    # Step 1.4:
    s = g1 ** n;
    z_plus = s;
    s_inv = s ** -1;
    z_minus = s_inv;

    i = 1;

    # Step 1.5:
    while True:

      timeout.check();

      # Step 1.5.1:
      T[z_plus] = i;
      T[z_minus] = -i;

      # Step 1.5.2:
      i = i + 1;
      if i > mpfr_ceil(B1 / n):
        # Step 1.5.2.1:
        break;

      # Step 1.5.3: Note: s and s^-1 were pre-computed above.
      z_plus = z_plus * s;
      z_minus = z_minus * s_inv;

    # Step 1.6:
    if verbose:
      print("The second stage begins:", \
              2 * B2 + 2 * (B2 + 1) * (n - 1), "operation(s)");

    z_plus = w;
    z_minus = w;
    j = 0;

    # Pre-compute group elements.
    g1_inv = g1 ** -1;
    g2_inv = g2 ** -1;

    g2_g1 = g2 * g1;
    g2_g1_inv = g2 * g1_inv;
    g2_inv_g1 = g2_inv * g1;
    g2_inv_g1_inv = g2_inv * g1_inv;

    # Step 1.7:
    while True:

      # Step 1.7.1:
      zp_plus = z_plus;
      zp_minus = z_minus;
      i = 0;

      # Step 1.7.2:
      while True:

        timeout.check();

        # Step 1.7.2.1:
        if zp_plus in T.keys():
          k = T[zp_plus];

          # Step 1.7.2.1.1:
          d = (nu1 + i - mpz(round(j * mu)) - k * n) * s1 + (nu2 + j) * s2;
          return int(d);

        # Step 1.7.2.2:
        if (j > 0) and (zp_minus in T.keys()):
          k = T[zp_minus];

          # Step 1.7.2.2.1:
          d = (nu1 + i - mpz(round(-j * mu)) - k * n) * s1 + (nu2 - j) * s2;
          return int(d);

        # Step 1.7.2.3:
        i = i + 1;
        if i >= n:
          # Step 1.7.2.3.1:
          break;

        # Step 1.7.2.4: Note: g1 was pre-computed above.
        zp_plus = zp_plus * g1;
        zp_minus = zp_minus * g1;

      # Step 1.7.3:
      j = j + 1;
      if j > B2:
        # Step 1.7.3.1:
        break;

      # Step 1.7.4:

      # Step 1.7.4.1:
      if round(j * mu) < round((j-1) * mu):

        # Step 1.7.4.1.1: Note: g2 g1 and g2^-1 g1^-1 were pre-computed above.
        z_plus = z_plus * g2_g1;
        z_minus = z_minus * g2_inv_g1_inv;

      # Step 1.7.4.2:
      elif round(j * mu) > round((j-1) * mu):

        # Step 1.7.4.2.1: Note: g2 g1^-1 and g2^-1 g1 were pre-computed above.
        z_plus = z_plus * g2_g1_inv;
        z_minus = z_minus * g2_inv_g1;

      # Step 1.7.4.3:
      else:

        # Step 1.7.4.3.1: Note: g2 and g2^-1 were pre-computed above.
        z_plus = z_plus * g2;
        z_minus = z_minus * g2_inv;

    # Step 1.8:
    return None;


  # Step 2: Setup the basis for the lattice.
  A = [[j, mpz(2 ** tau)], [mpz(2 ** (m + l)), mpz(0)]];


  # Step 3: Compute the reduced basis for the lattice and extract s1 and s2.
  [B, _] = lagrange(A);

  s1 = B[0];
  s2 = B[1];

  # Compute the norm lambda1 of s1.
  lambda1 = mpfr_sqrt((s1[0] ** 2) + (s1[1] ** 2));

  # Extra step: Verify the requirement on the norm.
  if None != t:
    if lambda1 < mpz(2 ** (m - t)):
      return None;

  # Step 4:

  # Compute mu, s2_parallel and then s2_perp given s2_parallel.
  mu = mpq(s1[0] * s2[0] + s1[1] * s2[1], s1[0] ** 2 + s1[1] ** 2);

  s2_parallel = [mu * s1[0], mu * s1[1]];
  s2_perp = [s2[0] - s2_parallel[0], s2[1] - s2_parallel[1]];

  # Compute the norm lambda2_perp of s2_perp.
  lambda2_perp = mpfr_sqrt(s2_perp[0] ** 2 + s2_perp[1] ** 2);


  # Step 5:

  # Define the vector v.
  v = [truncmod(-(2 ** m) * k, 2 ** (m + l)), mpz(0)];

  # Use Babai's nearest plane algorithm to find the vector o.
  o = babai(B, v);

  # Let nu1 and nu2 be integers such that o = nu1 * s1 + nu2 * s2.
  nu = solve_left(B, o);


  # Step 6:
  B1 = mpz(mpfr_floor((2 ** (m + tau)) * mpfr_sqrt(mpfr(2)) /
                        mpfr(lambda1) + 1));
  B2 = mpz(mpfr_floor((2 ** (m + tau)) * mpfr_sqrt(mpfr(2)) /
                        mpfr(lambda2_perp) + 1/2));

  if verbose:
    print("Computed B1 =", B1);
    print("Computed B2 =", B2);
    print("");


  # Step 7:
  return meet_in_the_middle(g,
                            x,
                            mpz(nu[0]),
                            mpz(nu[1]),
                            B1,
                            B2,
                            mpz(s1[1] // (2 ** tau)),
                            mpz(s2[1] // (2 ** tau)),
                            mu,
                            c);


def solve_multiple_j_k_for_d(
  j_k_list : list[list[int | mpz] | None],
  m : int,
  l : int,
  g : CyclicGroupElement,
  x : CyclicGroupElement,
  tau : int = 0,
  delta : float = LLL_DEFAULT_DELTA,
  precision : int | None = None,
  enumerate : bool | EnumerationOptions = False,
  timeout : int | None | Timeout = None,
  verbose : bool = False) -> int | mpz | None:

  """ @brief  Attempts to compute the short discrete logarithm d = log_g x given
              a list of n frequency pairs [[j_1, k_1], ..., [j_n, k_n]] yielded
              by n independent runs of the quantum part of Ekerå–Håstad's
              algorithm [EH17] by using the classical post-processing algorithm
              from [E20] and [E24t] (see Sect. 5.6).

      [EH17] Ekerå, M. and Håstad, J.: "Quantum Algorithms for Computing Short
                                        Discrete Logarithms and Factoring RSA
                                        Integers.". In: PQCrypto 2017.
                                       Springer LNCS 10346, pp. 347–363 (2017).

      [E20] Ekerå, M.: "On post-processing in the quantum algorithm for
                        computing short discrete logarithms".
                       Des. Codes Cryptogr. 88, pp. 2313–2335 (2020).

      [E24t] Ekerå, M.: "On factoring integers, and computing discrete
                         logarithms and orders, quantumly".
                        PhD thesis, KTH Royal Institute of Technology (2024).

      @param j_k_list   The n frequency pairs [[j_1, k_1], ..., [j_n, k_n]]
                        where j_1, ..., j_n are integers on [0, 2^(m + l)) and
                        k_1, ..., k_n are integers on [0, 2^l).

      @param m  A positive integer m such that d < 2^m.

      @param l  A positive integer l ≈ m / s for s a tradeoff factor. The two
                control registers in the quantum part of the algorithm are of
                lengths m + l and l qubits, respectively.

      @param g  The group element g.

      @param x  The group element x = g^d.

      @param tau  An integer tau on [0, l]. Used to scale the basis for the
                  lattice L^tau that is used in the post-processing, and that is
                  generated by the vector (j_1, ..., j_n, 2^tau), and by the n
                  vectors (2^(m + l), 0, ..., 0) thru (0, ..., 0, 2^(m + l), 0).

      @param delta  The parameter delta to use when delta-LLL-reducing the basis
                    for the lattice L^tau used in the post-processing. Must be
                    on the interval (1/4, 1]. A polynomial runtime in the
                    dimension of the lattice is only guaranteed for delta < 1.

      @param precision  The precision to use when computing the Gram–Schmidt
                        projection factors as a part of delta-LLL-reducing the
                        basis for the lattice L^tau used in the post-processing.

                        The precision may be set to None, as is the default, in
                        which case the projection factors are represented as
                        exact quotients.

      @param enumerate  A flag that may be set to True to enumerate vectors in
                        the lattice L^tau (until d is found or the specified
                        timeout has elapsed), or to EnumerationOptions.CVP to
                        consider only a closest vector in the lattice as
                        returned by performing a limited enumeration, or to
                        False to consider only the vector returned by Babai's
                        nearest plane algorithm.

                        May also be set to EnumerationOptions.BOUNDED_BY_TAU in
                        which case all vectors within distance R of the origin
                        of the lattice L^tau are enumerated, where R depends on
                        tau as R = sqrt(n + 1) * 2^(m + tau).

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the computation aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @param verbose  A flag that may be set to True to print intermediary
                      results and status updates when executing the
                      post-processing algorithm.

      @return   The logarithm d, or None, if solving for d fails. """

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  if None in j_k_list:
    return None;

  n = len(j_k_list);
  if n < 1:
    return None;

  js, ks = zip(*j_k_list)

  # Setup a basis for the lattice.
  if verbose:
    print("Setting up the basis...");

  A = [[mpz(j) for j in js] + [mpz(2 ** tau)]];
  for i in range(0, n):
    row = (n + 1) * [mpz(0)];
    row[i] = mpz(2 ** (m + l));
    A.append(row);

  # Reduce the basis for the lattice.
  if verbose:
    print("Computing the LLL-reduced basis with reduced precision...");

  B = lll(A,
          delta = delta,
          timeout = timeout,
          precision = LLL_DEFAULT_REDUCED_PRECISION);
            # Pre-compute with reduced precision.

  # Check the timeout.
  timeout.check();

  if verbose:
    print(" ** Refining the LLL-reduced basis with full precision...");

  [B, [Bs, M]] = lll(B,
                     delta = delta,
                     timeout = timeout,
                     gs = True,
                     precision = precision);

  # Check the timeout.
  timeout.check();

  # Setup v.
  if verbose:
    print("Setting up the target vector v...");

  v = [truncmod(-(2 ** m) * k, 2 ** (m + l)) for k in ks] + [mpz(0)];

  if enumerate in [EnumerationOptions.TRUE, True,
                   EnumerationOptions.FALSE, False,
                   EnumerationOptions.CVP]:

    # Find a vector u in the lattice that is close to v.
    if enumerate == EnumerationOptions.CVP:
      if verbose:
        print("Executing an enumeration to solve CVP...");

      # Let u be a closest vector to v in the lattice.
      u = solve_cvp(B, v, timeout = timeout, gs = [Bs, M]);
    else:
      if verbose:
        print("Executing Babai's nearest plane algorithm...");

      # Let u be a close vector to v in the lattice.
      u = babai(B, v, Bs);

    # Check if the correct solution was found.
    if verbose:
      print(" ** Checking:", u);

    candidate_d = u[-1];
    if 0 <= candidate_d < 2 ** (m + tau):
      candidate_d = candidate_d // mpz(2 ** tau);
      if g ** candidate_d == x:
        return candidate_d;

    if enumerate not in [True, EnumerationOptions.TRUE]:
      return None;

    # Check the timeout.
    timeout.check();

    if verbose:
      print("Enumerating vectors around v...");

    # Set the enumeration radius R = |u - v|^2.
    R2 = sum([(u[i] - v[i]) ** 2 for i in range(n + 1)]);

    # Enumerate with gradually increasing radii.
    while True:

      # Check the timeout.
      timeout.check();

      # Enumerate, whilst checking the timeout.
      for u in enumerate_lattice(B,
                                 R2,
                                 v,
                                 timeout = timeout,
                                 gs = [Bs, M]):

        # Check if the correct solution was found.
        if verbose:
          print(" ** Checking:", u);

        candidate_d = u[-1];
        if 0 <= candidate_d < 2 ** (m + tau):
          candidate_d = candidate_d // mpz(2 ** tau);
          if g ** candidate_d == x:
            return candidate_d;

      # Double the enumeration radius.
      if verbose:
          print(" ** Doubling the enumeration radius.");

      R2 = 4 * R2;

  elif enumerate == EnumerationOptions.BOUNDED_BY_TAU:

    # Enumerate, whilst checking the timeout.
    for u in enumerate_lattice(B,
                               mpz(n + 1) * mpz(2 ** (2 * (m + tau))),
                               v,
                               timeout = timeout,
                               gs = [Bs, M]):

      # Check if the correct solution was found.
      if verbose:
        print(" ** Checking:", u);

      candidate_d = u[-1];
      if 0 <= candidate_d < 2 ** (m + tau):
        candidate_d = candidate_d // mpz(2 ** tau);
        if g ** candidate_d == x:
          return candidate_d;

    # No solution was found.
    return None;

  else:
    raise Exception("Error: Incorrect parameters: Unknown enumerate option.");


""" @brief  A module for a utility class for collecting timing statistics. """

from __future__ import annotations;

from time import time;
from math import floor;

class Timer:

  """ @brief  A utility class for collecting timing statistics. """

  # A constant for state management indicating that the timer is stopped.
  STOPPED = 0;

  # A constant for state management indicating that the timer is running.
  RUNNING = 1;

  def __init__(self, delta_t = 0):

    """ @brief  Initializes the timer to a specific time delta. The timer is
                left in the stopped state until it is manually started.

        @param delta_t  The time delta. Defaults to zero. """

    self.state = Timer.STOPPED;
    self.delta_t = delta_t;

  def start(self) -> Timer:

    """ @brief  Starts the timer if it is currently stopped.

        @return   The timer. """

    if self.state == Timer.STOPPED:
      self.state = Timer.RUNNING;
      self.t = time();

    return self;

  def stop(self) -> Timer:

    """ @brief  Stops the timer if it is currently running.

        @return   The timer. """

    if self.state == Timer.RUNNING:
      self.state = Timer.STOPPED;
      self.delta_t += time() - self.t;

    return self;

  def reset(self, delta_t = 0) -> Timer:

    """ @brief  Stops the timer and re-initializes it to a specific time delta.

        @param delta_t  The time delta. Defaults to zero.

        @return The timer. """

    self.state = Timer.STOPPED;
    self.delta_t = delta_t;

    return self;

  def restart(self) -> Timer:

    """ @brief  Resets the timer and then starts it again.

        @return   The timer. """

    self.reset();
    self.start();

    return self;

  def peek(self) -> int | float:

    """ @brief  Peeks at the timer, returning the number of seconds elapsed.

        @return   If the timer is stopped, the time delta is returned.
                  Otherwise, the sum of the time delta and the current offset of
                  the timer returned. """

    tmp_delta_t = self.delta_t;
    if self.state == Timer.RUNNING:
      tmp_delta_t += time() - self.t;

    return tmp_delta_t;

  def __add__(a : Timer, b : Timer) -> Timer:

    """ @brief  Adds the time deltas of two stopped timers, returning a new
                timer with said time delta.

        The new timer is left in the stopped state until manually started.

        If either timer is running, an exception is raised.

        @param a   The first timer.

        @param b   The second timer.

        @return   A new timer, in the stopped state, with a time delta equal to
                  the sum of the time deltas of the first and second timers. """

    if (Timer.STOPPED != a.state) or (Timer.STOPPED != b.state):
      raise Exception("Error: Cannot add running timers.");

    return Timer(a.delta_t + b.delta_t);

  def __str__(self) -> str:

    """ @brief  Returns a string representation of the timer.

        @return   A string representation of the timer. """

    return Timer.format(self.peek());

  def __repr__(self) -> str:

    """ @brief  Returns a string representation of the timer.

        @return   A string representation of the timer. """

    return str(self);

  @staticmethod
  def format(delta_t : int | float) -> str:

    """ @brief  Formats a time delta in seconds as a string.

        @param delta_t  The time delta in seconds.

        @return   A string representation of the time delta. """

    # Compute days, hours, minutes, seconds, milliseconds and microseconds.
    days = floor(delta_t / 3600 / 24);
    hours = floor(delta_t / 3600) % 24;
    mins = floor(delta_t / 60) % 60;
    secs = floor(delta_t) % 60;
    ms = int(floor((10 ** 3) * delta_t)) % (10 ** 3);
    us = int(floor((10 ** 6) * delta_t)) % (10 ** 3);

    # Format the time delta as a human-readable string.
    hr = "";

    if days > 0:
      if hr != "":
        hr += " "
      hr += str(days) + " d";

    if hours > 0:
      if hr != "":
        hr += " "
      hr += str(hours) + " h";

    if mins > 0:
      if hr != "":
        hr += " "
      hr += str(mins) + " m";

    if secs > 0:
      if hr != "":
        hr += " "
      hr += str(secs) + " s";

    if ms > 0:
      if hr != "":
        hr += " "
      hr += str(ms) + " ms";

    if us > 0:
      if hr != "":
        hr += " "
      hr += str(us) + " µs";

    # Handle the special case of the time delta being zero.
    if hr == "":
      if isinstance(delta_t, int):
        hr = "0 s";
      else:
        hr = "0 µs";

    return hr;


""" @brief  A module for collecting the non-trivial factors of N. """

from gmpy2 import iroot;
from gmpy2 import is_prime;

from gmpy2 import mpz;
from gmpy2 import gcd;

from ......utils.timer import Timer;

class FactorCollection:

  """ @brief  A class for collecting the non-trivial factors of N. """

  def __init__(self, N):

    """ @brief  Initializes the collection of non-trivial factors of N.

        @param N  The integer N. """

    # The integer N to be factored.
    self.N = mpz(N);

    # The set of factors found thus far, reduced so that all factors in the set
    # are pairwise coprime to each other. This property is enforced by add().
    self.found_factors = set();

    # The set of prime factors found thus far; a subset of found_factors.
    self.found_primes = set();

    # A timer for measuring the time spent performing primality tests.
    self.timer_test_primality = Timer();

    # A timer for measuring the time spent detecting perfect powers.
    self.timer_test_perfect_power = Timer();

    # The residual; the product of the composite pairwise coprime factors in the
    # collection, or one if there are no composite factors in the collection.
    self.residual = 1;

    # Add N as a factor.
    self.add(N);

  def is_complete(self):

    """ @brief  Checks if all prime factors of N have been found.

        @return True if all prime factors have been found, False otherwise. """

    return self.residual == 1;

  def add(self, d):

    """ @brief  Adds a factor d to this collection.

        @param d  The factor to add to this collection. """

    d = mpz(d);

    # Check that the factor is non-trivial and has not already been found.
    if (d == 1) or (d in self.found_factors):
      return;

    # Test if d shares a factor with any of the factors found.
    D = mpz(1);

    for f in self.found_factors:
      D = gcd(f, d);

      if D != 1:
        break;

    if D != 1:
      # If so, remove f, split f and d, and add the resulting factors.
      self.found_factors.remove(f);
      if f not in self.found_primes:
        # Also remove f from the residual when removing f from the collection.
        self.residual //= f;

      f //= D;
      d //= D;

      self.add(D);

      if f != 1:
        self.add(f);

      if d != 1:
        self.add(d);
    else:
      # Check if d is a perfect power, and if so reduce d.
      self.timer_test_perfect_power.start();

      if d.is_power():
        for e in range(2, d.bit_length()):
          (d_reduced, result) = iroot(d, e);
          if result:
            d = d_reduced;
            break;

      self.timer_test_perfect_power.stop();

      # Add d to the factors found.
      self.found_factors.add(d);

      # Check if d is prime, and if so register it.
      self.timer_test_primality.start();
      result = is_prime(d);
      self.timer_test_primality.stop();

      if result:
        self.found_primes.add(d);
      else:
        # If d is not prime, multiply d onto the residual.
        self.residual *= d;

  def print_status(self):

    """ @brief  Prints status information for this collection. """

    print("Found factors:", len(self.found_factors));
    print("Found primes:", len(self.found_primes));

    found_factors = list(self.found_factors);
    found_factors.sort();

    for i in range(len(found_factors)):
      print(" Factor " + str(i) + ":", found_factors[i]);
    print("");

  def __eq__(self, other):

    """ @brief  Compares this collection to another collection.

        @param other  The other collection to which to compare this collection.

        @return   True if the collections are equal, False otherwise. """

    if self.N != other.N:
      return False;
    if self.residual != other.residual:
      return False;
    if self.found_factors != other.found_factors:
      return False;
    if self.found_primes != other.found_primes:
      return False;

    return True; # Note: Disregards timer differences.

  def __str__(self):

    """ @brief  Returns a string representation of this collection.

        @return   A string representation of this collection. """

    return str(self.found_factors);

  def __repr__(self):

    """ @brief  Returns a string representation of this collection.

        @return   A string representation of this collection. """

    return str(self);


""" @brief  A module for Babai's nearest plane algorithm. """

import gmpy2;

from gmpy2 import mpz;
from gmpy2 import mpq;
from gmpy2 import mpfr;

from .gram_schmidt import gram_schmidt;
from ..matrices import dimensions;

def babai(
  B : list[list[int | mpz]],
  t : list[int | mpz | mpq],
  Bs : list[list[int | mpz | mpq | mpfr]] | None = None,
  precision : int | None = None) -> list[int | mpz]:

  """ @brief  Uses Babai's nearest plane algorithm [Babai86] to find a vector x
              that is in the lattice L, and that is close to a target vector t,
              where L is generated by the rows of a reduced basis matrix B.

      More specifically, it holds that | t - x | <= 2^(n/2) dist(t, L) if B is
      delta-LLL-reduced with delta >= 3/4.

      [Babai86] Babai, L.: "On Lovász’ lattice reduction and the nearest lattice
                            point problem". Combinatorica 6(1), pp. 1–13 (1986).

      The basis matrix B is represented as as list of lists [b_1, ..., b_n],
      where b_i = [b_i1, ..., b_id] for i = 1, ..., n represent the n row
      vectors that make up the basis.

      @param B  The n x d basis basis matrix B = [b_1, ..., b_n], where
                b_i = [b_i1, ..., b_id] for i = 1, ..., n represent the n row
                vectors that make up the basis.

      @param t  The target vector t = [t_1, ..., t_d].

      @param Bs   The Gram–Schmidt orthogonalization Bs of B, if available, or
                  None in which case the Gram–Schmidt orthogonalization B is
                  computed by this function. (If you plan on calling this
                  function several times for the same lattice L, then you may
                  wish to pre-compute Bs and pass it along in each call to this
                  function. Note also that the function lll() for computing
                  delta-LLL-reduced bases B optionally computes Bs alongside B.)

      @param precision  The precision to use when computing the Gram–Schmidt
                        projection factors in M. May be set to None, in which
                        case projection factors are represented as exact
                        quotients.

                        Note that this parameter only has an effect if gs is
                        set to None as the Gram–Schmidt orthogonalized basis Bs
                        and the associated matrix M of Gram–Schmidt projection
                        factors are otherwise pre-computed.

      @return   A vector x that is in the lattice L generated by B, and that is
                close to t. More specifically, | t - x | <= 2^(n/2) dist(t, L)
                if B is delta-LLL-reduced with delta >= 3/4. """

  # Initial setup.
  (n, d) = dimensions(B);

  if None == Bs:
    [Bs, _] = gram_schmidt(B, precision = precision);

  with gmpy2.context(gmpy2.get_context()) as context:

    # Set the precision.
    if precision != None:
      context.precision = precision;

    # Step 2.
    x = t.copy();

    # Step 3.
    for i in range(n, 0, -1):
      # Step 3.1.
      a = sum([x[j] * Bs[i-1][j] for j in range(d)]);
      b = sum([Bs[i-1][j] * Bs[i-1][j] for j in range(d)]);

      if None == precision:
        mu = mpq(a, b);
      else:
        mu = mpfr(a) / mpfr(b);

      # Step 3.2.
      mu_rounded = round(mu);

      for j in range(d):
        x[j] = x[j] - mu_rounded * B[i-1][j];

    # Step 4.
    for j in range(d):
      x[j] = t[j] - x[j];

    # Step 5.
    return x;


""" @brief  The root module for the Quaspy (Quantum algorithm simulations in
            Python) library for Python.

    Quaspy contains modules that implement:

    - Simulators for the quantum part of Shor's order-finding algorithm
      [Shor94], modified as in [E24], and the classical post-processing from
      [E24] that recovers the order in a single run with very high success
      probability. Furthermore, these modules implement simulators for the
      quantum part of Seifert's variation [Seifert01] of Shor's order-finding
      algorithm [Shor94], as described as in [E24t] and [E21], and with the
      classical post-processing from [E24t] and [E21] with supporting functions
      from [E24] that efficiently recovers the order, or a positive integer
      multiple of the order, in multiple runs when making tradeoffs.

    - Simulators for factoring general integers via order-finding, and the
      classical post-processing from [E21b] and [E24] that factors any integer
      completely in a single order-finding run with very high success
      probability.

    - Simulators for the quantum part of Shor's algorithm for computing general
      discrete logarithms [Shor94], modified as in [E19p], and the classical
      post-processing from [E19p] that recovers the logarithm given the order in
      a single run with very high probability of success, and that also
      efficiently recovers the logarithm given the order when making tradeoffs.

    - Simulators for the quantum part of Ekerå–Håstad's algorithm for computing
      short discrete logarithms [EH17], modified as in [E20] and [E23p], and the
      classical post-processing from [E23p] that recovers the logarithm in a
      single run with very high probability of success. This algorithm does not
      require the order to be known. Furthermore, these modules implement the
      classical post-processing from [E20] that efficiently recovers the
      logarithm in multiple runs of the quantum part of Ekerå–Håstad's algorithm
      when making tradeoffs.

    - Simulators for factoring RSA integers via short discrete logarithms, by
      using the reduction in [EH17], modified as in [E20] and [E23p], and the
      classical post-processing from [E23p] that factors random RSA integers in
      a single run of the quantum part of Ekerå–Håstad's algorithm with very
      high probability of success. Furthermore, these modules implement the
      classical post-processing from [E20] that efficiently factors random RSA
      integers in multiple runs of the quantum part of Ekerå–Håstad's algorithm
      when making tradeoffs.

    All modules, classes, methods and functions in Quaspy are documented using
    Python docstrings (see https://github.com/ekera/quaspy/blob/main/docs).

    Note that Quaspy implements basic support for tradeoffs via a native Python
    implementation of LLL that is stable and reasonably efficient. See also the
    Qunundrum repository (see https://github.com/ekera/qunundrum) with its suite
    of MPI programs that implements support for tradeoffs via LLL and BKZ as
    implemented by fpLLL (see https://github.com/fplll/fplll).

    Note furthermore that portions of Quaspy are inherited from the Factoritall
    repository (see https://github.com/ekera/factoritall).

    Quaspy is a work in progress, and may be subject to major changes without
    prior notice. Quaspy was developed for academic research purposes. It grew
    out of our research project in an organic manner as research questions were
    posed and answered. It is distributed "as is" without warranty of any kind,
    either expressed or implied. For further details, see the license (see
    https://github.com/ekera/quaspy/blob/main/LICENSE.md).

    Quaspy was developed by Martin Ekerå, in part at KTH, the Royal Institute of
    Technology, in Stockholm, Sweden. Valuable comments and advice were provided
    by Johan Håstad throughout the development process. Funding and support was
    provided by the Swedish NCSA that is a part of the Swedish Armed Forces.

    For further details on Quaspy, see the Quaspy repository on GitHub
    (available at https://github.com/ekera/quaspy).

    [Shor94] Shor, P.W.: "Algorithms for Quantum Computation: Discrete
                          Logarithms and Factoring".
                         In: Proceedings from FOCS '94, pp. 124–134 (1994).

    [Seifert01] Seifert, J.-P.: "Using fewer qubits in Shor's factorization
                                 algorithm via simultaneous Diophantine
                                 approximation". In: CT-RSA 2001.
                                Springer LNCS 2020, pp. 319–227 (2001).

    [EH17] Ekerå, M. and Håstad, J.: "Quantum Algorithms for Computing Short
                                      Discrete Logarithms and Factoring RSA
                                      Integers.". In: PQCrypto 2017.
                                     Springer LNCS 10346, pp. 347–363 (2017).

    [E19p] Ekerå, M.: "Revisiting Shor's quantum algorithm for computing
                       general discrete logarithms".
                      ArXiv 1905.09084v4 (2024).

    [E20] Ekerå, M.: "On post-processing in the quantum algorithm for
                      computing short discrete logarithms".
                     Des. Codes Cryptogr. 88, pp. 2313–2335 (2020).

    [E21b] Ekerå, M.: "On completely factoring any integer efficiently in a
                       single run of an order-finding algorithm".
                      Quantum Inf. Process. 20(6):205 (2021).

    [E23p] Ekerå, M.: "On the success probability of the quantum algorithm for
                       the short DLP". ArXiv 2309.01754v2 (2025).

    [E24] Ekerå, M.: "On the success probability of quantum order finding".
                     ACM Trans. Quantum Comput. 5(2):11 (2024).

    [E24t] Ekerå, M.: "On factoring integers, and computing discrete logarithms
                       and orders, quantumly".
                      PhD thesis, KTH Royal Institute of Technology (2024). """

# Shor's algorithm for splitting N if r is even and g^(r/2) != -1 (mod N).
from .factoring.general.postprocessing.shor import split_N_given_g_r;

# The algorithm from [E21b] for completely factoring N given r.
from .factoring.general.postprocessing.ekera import solve_r_for_factors;

# Convenience functions implementing algorithms from [E21b] and [E24] for first
# solving j for r, and then solving r and N for the complete factorization of N.
from .factoring.general.postprocessing.ekera import solve_j_for_factors;
from .factoring.general.postprocessing.ekera import solve_j_for_factors_mod_N;

# Convenience function implementing the algorithm from [EH17], with improvements
# from [E20], for splitting N given d by solving a quadratic equation.
from .factoring.rsa.postprocessing import split_N_given_d;

# Convenience function implementing the algorithm from [EH17], with improvements
# from [E20], for setting up x = g^d given g and N = pq, and for computing d
# given the primes p and q.
from .factoring.rsa import setup_d_given_p_q;
from .factoring.rsa import setup_x_given_g_N;

# The algorithms in [E21b], [E24t] (see in particular Sect. 5.2.3) and the
# Factoritall repository (available at https://github.com/ekera/factoritall) for
# sampling g uniformly at random from the multiplicative group of the ring of
# integers modulo N, and for returning [g, r] or r, for r the (in some cases
# heuristically computed) order of g.
from .factoring.sampling import sample_g_r_given_N;
from .factoring.sampling import sample_r_given_N;

# The post-processing algorithms in [E24] for solving j for r.
from .orderfinding.general.postprocessing.ekera import solve_j_for_r;
from .orderfinding.general.postprocessing.ekera import solve_j_for_r_mod_N;

# Algorithms for sampling j given r to test the above order-finding algorithms.
from .orderfinding.general.sampling import sample_j_given_r;

# The post-processing algorithm from [E19p] for solving (j, k) for d given r.
from .logarithmfinding.general.postprocessing import solve_j_k_for_d_given_r;

# Algorithms for heuristically sampling (j, k) given d and r to test the above
# post-processing algorithm.
from .logarithmfinding.general.sampling import sample_j_k_given_d_r_heuristic;

# The post-processing algorithm from [E23p] for solving (j, k) for d.
from .logarithmfinding.short.postprocessing import solve_j_k_for_d;

# Algorithms for heuristically sampling (j, k) given d, r and tau to test the
# above post-processing algorithm.
from .logarithmfinding.short.sampling import sample_j_k_given_d_r_tau;



""" @brief  A module for internal classes and functions used by the functions in
            the parent module for solving a frequency j yielded by the quantum
            part of Shor's order-finding algorithm for the order r. """


""" @brief  A module for solving an optimal frequency j = j0(z) for
            r_tilde = r / d where d = gcd(r, z). """

import gmpy2;

from gmpy2 import mpfr;
from gmpy2 import mpz;

from gmpy2 import gcd;

from math import sqrt;
from math import floor;

from .collection import CandidateCollection;

from ......math.continued_fractions import continued_fractions;
from ......math.lattices.lagrange import lagrange;
from ......math.vectors import norm2;

from ......math.primes import prime_power_product;

def solve_j_for_r_tilde_continued_fractions(j, m, l):

  """ @brief  For j = j0(z) an optimal frequency, for m such that r < 2^m and l
              such that r^2 < 2^(m+l), and any z in [0, r), this function
              recovers r_tilde = r / d where d = gcd(r, z) by expanding the
              quotient j / 2^(m+l) in continued fractions and returning the
              last denominator < 2^((m+l)/2) as described in [E24].

      [E24] Ekerå, M.: "On the success probability of quantum order finding".
                       ACM Trans. Quantum Comput. 5(2):11 (2024).

      By Lem. 4.1 in [E24], this function is guaranteed to return r_tilde
      provided that the requirements on the input parameters are met.

      For further details, see Lem. 4.1, and Sect. 4 and App. B, of [E24].

      @param j  An optimal frequency j0(z), for m and l as passed to this
                function, and for any z in [0, r).

                If some other frequency j on [0, 2^(m+l)) is passed to this
                function, it may or may not return r_tilde.

      @param m  A positive integer m such that r < 2^m.

      @param l  A positive integer l <= m, such that r^2 < 2^(m+l).

      @return   The last denominator < 2^((m+l)/2) in the continued fraction
                expansion of j / 2^(m+l). This denominator is guaranteed to be
                equal to r_tilde, provided that the requirements on the input
                parameters are met. """

  bound = mpfr(2) ** mpfr((m + l) / 2);

  r_tilde_candidates = continued_fractions(j, m, l, denominator_bound = bound);

  return r_tilde_candidates[-1];


def solve_j_for_r_tilde_lattice_svp(j, m, l, multiples = None):

  """ @brief  For j = j0(z) an optimal frequency, for m such that r < 2^m and l
              such that r^2 < 2^(m+l), and any z in [0, r), this function
              recovers r_tilde = r / d where d = gcd(r, z) by finding the
              shortest non-zero vector in a two-dimensional lattice L as
              described in [E24].

      [E24] Ekerå, M.: "On the success probability of quantum order finding".
                       ACM Trans. Quantum Comput. 5(2):11 (2024).

      More specifically, this function uses Lagrange's lattice basis reduction
      algorithm to find the shortest non-zero vector

        u = (rj - 2^(m+l) z, r / 2) / d

      in the lattice L generated by (j, 1/2) and (2^(m+l), 0), and hence
      r_tilde, as the second component is r_tilde / 2. This function returns
      r_tilde.

      By Lem. 4.2 in [E24], provided that r < 2^m and r^2 < 2^(m+l), the second
      component of the shortest non-zero vector in L has r_tilde / 2 in its
      second component, up to sign of course.

      For further details, see Lem. 4.2, and Sect. 4 and App. C, of [E24].

      @param j  An optimal frequency j0(z), for m and l as passed to this
                function, and for any z in [0, r).

                If some other frequency j on [0, 2^(m+l)) is passed to this
                function, it will also return an integer, that may or may not
                be equal to r_tilde.

      @param m  A positive integer m such that r < 2^m.

      @param l  A positive integer l <= m, such that r^2 < 2^(m+l).

      @param multiples  Row multiples to use as a starting point when computing
                        the Lagrange-reduced basis for the slightly scaled basis
                        A = [[j, 1], [2^(m+l+1), 0]], see the lagrange()
                        function for further details, or None, as is the
                        default, if no such multiples are available.

                        The idea is that when trying to solve not only j but
                        also j ± 1, ..., j ± B for r_tilde, in the hope that
                        this will lead to an optimal frequency j0(z) being
                        solved for r_tilde, the row multiples that yield a
                        Lagrange-reduced basis for adjecent offsets in j are
                        likely to be close.

                        For this reason, this function accepts row multiples as
                        input, and it furthermore returns as output the row
                        multiples that yield a Lagrange-reduced basis for the
                        value of j input, so that these can be fed back to this
                        function when processing j + 1 or j - 1, recursively.

      @return   The tuple [r_tilde_candidate, multiple'], where multiple' are
                the row multiples that yield a reduced basis for the value of
                j input, and r_tilde_candidate is equal to r_tilde provided that
                the requirements on the input parameters are met. """

  # Setup the basis matrix for the lattice L (scaled by a factor of two).
  A = [[mpz(2 * j), mpz(1)], [2 * (2 ** (m + l)), mpz(0)]];
  # Without scaling: A = [[mpz(offset_j), mpq(1, 2)], [pow2ml, mpz(0)]];

  # Reduce the basis matrix.
  [A, multiples] = lagrange(A, multiples = multiples);

  # Extract r_tilde from the reduced basis (scaled by a factor of two).
  r_tilde_candidate = abs(mpz(A[0][1]));
  # Without scaling: r_tilde_candidate = abs(mpz(2 * A[0][1]));

  return [r_tilde_candidate, multiples];


def solve_j_for_r_tilde_lattice_enumerate(
  j,
  m,
  l,
  g,
  g_pow_e_context = None,
  c = 1,
  accept_multiple = False,
  partial_exponentiation = True,
  filtered_r_tilde_candidates = None,
  dismissed_reduced_r_tilde_candidates = None,
  mu = 0,
  multiples = None,
  verbose = False):

  """ @brief  For j = j0(z) an optimal frequency, for m such that r < 2^m and
              l = m - Delta for some Delta in [0, m), and z in [0, r), this
              function recovers r_tilde = r / d where d = gcd(r, z), provided
              that d is cm-smooth, by enumerating at most 6 * sqrt(3) * 2^Delta)
              vectors in a two-dimensional lattice L as described in [E24].

      [E24] Ekerå, M.: "On the success probability of quantum order finding".
                       ACM Trans. Quantum Comput. 5(2):11 (2024).

      More specifically, this function uses Lagrange's lattice basis reduction
      algorithm to find a reduced basis for the lattice L generated by (j, 1/2)
      and (2^(m+l), 0). It then enumerates all vectors within a ball of radius
      2^(m-1/2) in L centered at the origin to find

        u = (rj - 2^(m+l) z, r / 2) / d,

      and hence r_tilde, as the second component is r_tilde / 2.

      By Lem. 4.3 in [E24], provided that r < 2^m and l = m - Delta, at most
      6 * sqrt(3) * 2^Delta vectors must be enumerated in L to find u and hence
      r_tilde, so if Delta is small then this method is efficient.

      In practice, as mentioned in [E24], the leading constant in the above
      bound is not tight, and the enumeration can be optimized. Some of these
      optimizations have been integrated into this implementation, so the
      enumeration typically considers fewer points than the bound indicates.

      @remark   Unlike the solve_j_for_r_tilde_lattice_svp() function, and the
                solve_j_for_r_tilde_continued_fractions(), this function does
                check that the candidates x for r_tilde returned fulfill the
                requirement that e(x) * x is a positive multiple of r, where
                e(x) is cm-smooth by the definition of cm-smooth in [E24].

                This is necessary as the enumeration may generate a
                comparatively, and it can also be done efficiently.

                A side effect of this test being performed is that this function
                will typically not find r_tilde = r / d if d is not cm-smooth.

      @remark   By default, this implementation uses an instance of the
                CandidateCollection to keep track of the collection of
                candidates for r_tilde found during the enumeration.

                As is explained in the documentation for the CandidateCollection
                class, this class keeps the collection reduced. In turn, this
                may lead to r_tilde not being returned as one of the candidates
                for r_tilde that represent the collection. This being said, one
                of the candidates x for r_tilde that represent the collection
                will then be such that r = e(x) * x, where e(x) is cm-smooth,
                and from such x it is always possible to correctly recover r.

                As is also explained below, you may input set() instead of
                CandidateCollection() to the filtered_r_tilde_candidates
                parameter to ensure all candidates for r_tilde are returned.

      For further details, see Lem. 4.3, and Sect. 4 and App. C, of [E24].

      @param j  An optimal frequency j0(z), for m and l as passed to this
                function, and for z in [0, r).

      @param m  A positive integer m such that r < 2^m.

      @param l  A positive integer l <= m, such that m + l is the length of the
                control register in the quantum order-finding algorithm.

                As is explained in [E24], it is possible to select l = m-Delta,
                for some Delta in [0, m), at the expense of enumerating at most
                6 * sqrt(3) * 2^Delta lattice vectors.

      @param g  The group element g of order r.

      @param g_pow_e_context  The pair [x, e] for x = g^e and e a maximal
                              cm-smooth product, or None, in which case this
                              function will compute x and e.

      @param c  A parameter c >= 1 that specifies the maximum size of the
                missing cm-smooth component d in r = d * r_tilde. As is
                explained in [E24], increasing c increases the success
                probability, at the expense of increasing the runtime.

      @param accept_multiple  A flag that may be set to True to indicate that
                              only a positive integer multiple of r is sought.
                              If set to True, this function returns as soon as
                              it finds a candidate x for r_tilde such that
                              g^(d * x) = 1, for d a cm-smooth prime power
                              product.

      @param partial_exponentiation   A flag that may be set to True to indicate
                                      that x^s1[1] and x^s1[2] shall first be
                                      computed, and that the candidate for
                                      r_tilde for element i1 * s[1] + i2 * s[2]
                                      in the lattice shall then be computed by
                                      as (x1^i1) * (x2^i2).

      @param filtered_r_tilde_candidates  A collection of candidates x for
                                          r_tilde, such that g^(e(x) * x) = 1,
                                          for e(x) some cm-smooth prime power
                                          product, and such that x is in the
                                          integer range [1, 2^m).

                                          When e.g. solving not only j, but also
                                          j ± 1, ..., j ± B, for r, for B some
                                          bound, then this parameter allows for
                                          accounting for candidates of r_tilde
                                          found in enumerations for previous
                                          offsets when enumerating for a given
                                          offset. In particular, it removes the
                                          need to perform exponentiations to
                                          test candidates x for r_tilde that
                                          have been found before, and checked
                                          to meet the requirement that
                                          g^(e(x) * x) = 1, for e(x) some
                                          cm-smooth prime power product.

                                          May be set to CandidateCollection(),
                                          as is the default when passing None,
                                          to start with an empty collection of
                                          candidates for r_tilde that is kept
                                          reduced, or to set(), to keep all
                                          candidates for r_tilde found during
                                          the enumeration in the set.

                                          The set of candidates for r_tilde
                                          returned by this function will be of
                                          the same type as the set input via
                                          this parameter.

      @param dismissed_reduced_r_tilde_candidates   A set of reduced candidates
                                                    for r_tilde that have
                                                    already been dismissed and
                                                    may be dismissed immediately
                                                    if encountered again.

                                                    May be set to set(), as is
                                                    the default when passing
                                                    None, if no candidates for
                                                    r_tilde have been dismissed.

      @param mu   A non-negative multiple of the order r used to reduce the
                  candidates for r_tilde found during the enumeration. May be
                  set to zero, as is the default, if no non-trivial multiple of
                  r is known.

      @param multiples  Row multiples to use as a starting point when computing
                        the Lagrange-reduced basis for the slightly scaled basis
                        A = [[j, 1], [2^(m+l+1), 0]], see the lagrange()
                        function for further details, or None, as is the
                        default, if no such multiples are available.

                        The idea is that when trying to solve not only j but
                        also j ± 1, ..., j ± B for r_tilde, in the hope that
                        this will lead to an optimal frequency j0(z) being
                        solved for r_tilde, the row multiples that yield a
                        Lagrange-reduced basis for adjecent offsets in j are
                        likely to be close.

                        For this reason, this function accepts row multiples as
                        input, and it furthermore returns as output the row
                        multiples that yield a Lagrange-reduced basis for the
                        value of j input, so that these can be fed back to this
                        function when processing j + 1 or j - 1, recursively.

      @param verbose  A flag that may be set to True to print intermediary
                      results and status updates when performing th
                      enumeration.

      @return   The tuple [filtered_r_tilde_candidates', [success,
                dismissed_reduced_r_tilde_candidates', mu', multiples']], where

                - filtered_r_tilde_candidates' is filtered_r_tilde_candidates
                  as input to this function updated with the additional
                  candidates for r_tilde found when enumerating the lattice
                  (these candidates have all passed the filtration step, and are
                   such that d * t is a multiple of r, where d is cm-smooth),

                - success is set to True if at least one candidate for r_tilde
                  passed the filter when enumerating the lattice, even if this
                  candidate was already in filtered_r_tilde_candidates, and
                  to False otherwise,

                - dismissed_reduced_r_tilde_candidates' is
                  dismissed_reduced_r_tilde_candidates as input to this function
                  updated with the additional reduced candidates for r_tilde
                  dismissed when enumerating the lattice,

                - mu' is an updated non-negative multiple of r, obtained by
                  letting mu = mu' and then updating mu' recursively by letting
                  mu' = gcd(mu', filted_r_tilde_candidate * e), where e is a
                  maximal cm-smooth product and x runs over all candidates for
                  r_tilde that pass the filtration step when enumerating the
                  lattice, and

                - multiple' are the row multiples that yield a reduced basis for
                  the value of j input. """


  # Enforce default. Note: We cannot use CandidateCollection() or set() as
  # defaults in the prototype because Python caches these collections/sets.
  if None == filtered_r_tilde_candidates:
    filtered_r_tilde_candidates = CandidateCollection();

  if None == dismissed_reduced_r_tilde_candidates:
    dismissed_reduced_r_tilde_candidates = set();

  # Sanity checks.
  if (m <= 0) or (l < 0) or (l > m):
    raise Exception("Error: Incorrect parameters");

  with gmpy2.context(gmpy2.get_context()) as context:

    # Set the precision.
    context.precision = 53;

    # Compute Delta given m and l.
    Delta = m - l;

    # Setup the basis matrix for the lattice L (scaled by a factor of two).
    A = [[mpz(2 * j), mpz(1)], [2 * mpz(2 ** (m + l)), mpz(0)]];
    # Without scaling: A = [[mpz(j), mpq(1, 2)], [mpz(2 ** (m + l), mpz(0)]];

    # Reduce the basis matrix.
    [A, multiples] = lagrange(A, multiples = multiples);

    # Extract the shortest non-zero vector, denoted s1, and the shortest
    # non-zero vector that is linerly independent to s1, denoted s2.
    [s1, s2] = A;

    # Compute float representation of these vectors, since they may be large.
    s1f = [mpfr(x) for x in s1];
    s2f = [mpfr(x) for x in s2];

    # Compute the Gram–Schmidt-coefficient mu21, such that
    #   mu12 * s1 = component of s2 that is parallel to s1, and
    #   s2 - mu12 * s1 = component of s2 that is orthogonal to s1.
    mu12 = (s1f[0] * s2f[0] + s1f[1] * s2f[1]) / norm2(s1f);

    # Compute the parallel and orthogonal components of s2.
    s2f_parallel = [mu12 * s1f[0], mu12 * s1f[1]];
    s2f_orthogonal = [s2f[0] - s2f_parallel[0], s2f[1] - s2f_parallel[1]];

    if None == g_pow_e_context:
      # Form e.
      e = prime_power_product(c * m);

      # Exponentiate g to e to form x.
      x = g ** e;
    else:
      [x, e] = g_pow_e_context;

      e = mpz(e);

    # The radius of the circle to enumerate. In [E24], the radius of the circle
    # to enumerate is 2^(m - 1/2), which would imply radius2 = 2^(2m - 1). This
    # bound stems from the fact that the target vector is
    #
    #   | [alpha_0(z) / d, (r / 2) / d] | <= | [r/2, r/2] | =
    #     sqrt((r/2)^2 + (r/2)^2) = sqrt(r^2 / 2) = r / sqrt(2) <
    #       2^m / sqrt(2) = 2^(m-1/2),
    #
    # since d = gcd(z, r) >= 1, |alpha_0(z)| <= r/2 and r < 2^m.
    #
    # We have scaled the lattice by a factor of two, yielding instead
    #
    #   | [2 alpha_0(z) / d, r / d] | <= | [r, r] | =
    #     sqrt(r^2 + r^2) = sqrt(2 r^2) = sqrt(2) r <
    #       sqrt(2) 2^m = 2^(m+1/2),
    #
    # and (2^(m+1/2))^2 = 2^(2m+1) as below.
    radius2 = mpfr(mpz(2 ** (2 * m + 1)));

    # A flag that is set to True if a candidate was found, otherwise to False.
    success = False;

    # Setup mu.
    mu = mpz(mu);

    if norm2(s2f_orthogonal) >= radius2:
      # As is stated in [E24], if | s2_orthogonal | >= radius^2, we have that
      # the second component of the shortest non-zero vector must be r_tilde / 2
      # (and we have scaled by factor of two, so the component is now r_tilde).

      r_tilde_candidate = abs(s1[1]);

      if (r_tilde_candidate >= 1) and (r_tilde_candidate < (2 ** m)):
        if r_tilde_candidate in filtered_r_tilde_candidates:
          success = True;
        else:
          # Use that mu is an r-multiple to reduce the candidate for r_tilde.
          reduced_r_tilde_candidate = gcd(r_tilde_candidate, mu);

          if (reduced_r_tilde_candidate in \
              dismissed_reduced_r_tilde_candidates):
            # Dismiss the reduced candidate.
            if verbose:
              print("Dismissing:", r_tilde_candidate);
          else:
            # The reduced candidate has not already been dismissed.
            if verbose:
              print("Testing the reduced candidate:", \
                reduced_r_tilde_candidate);

            # Test the reduced candidate.
            if (x ** reduced_r_tilde_candidate) == 1:
              success = True;

              # Add r_tilde_candidate to the filtered candidates for r_tilde.
              filtered_r_tilde_candidates.add(r_tilde_candidate);

              # We know that reduced_r_tilde_candidate * e is a multiple of r,
              # so we may update mu to account for this fact:
              mu = gcd(reduced_r_tilde_candidate * e, mu);
            else:
              # Add reduced_r_tilde_candidate to the dismissed reduced
              # candidates for r_tilde to avoid future exponentiations.
              dismissed_reduced_r_tilde_candidates.\
                add(reduced_r_tilde_candidate);

      return [filtered_r_tilde_candidates,
              [success,
              dismissed_reduced_r_tilde_candidates,
              mu,
              multiples]];

    # Compute an upper bound B on the number of points to enumerate.
    B = floor(6 * sqrt(3) * (2 ** Delta));

    # Pre-compute 2^m for later comparisons.
    pow2mf = mpfr(mpz(2 ** m));

    # Storage for x_basis = [x ** s1[1], x ** s2[1]] that is precomputed upon
    # demand if the partial_exponentiation flag is set to True.
    x_basis = None;

    count = 0;

    i2 = 0;

    while True:
      # Check the condition on the radius.
      u2_orthogonalf = [i2 * s2f_orthogonal[0], i2 * s2f_orthogonal[1]];
      if norm2(u2_orthogonalf) > radius2:
        break;

      # Form u2f.
      u2f = [i2 * s2f[0], i2 * s2f[1]];

      # Form i1hat and search i1 = i1hat, i1hat ± 1, i1hat ± 2, ..
      i1 = i1hat = round(-mu12 * i2);

      # Optimization: Jump ahead in i1.
      uf = [u2f[0] + i1 * s1f[0], u2f[1] + i1 * s1f[1]];

      if s1f[1] >= 0:
        if uf[1] <= -pow2mf:
          i1 += int(floor((-uf[1] - pow2mf) // s1f[1]));
      else:
        if uf[1] >= pow2mf:
          i1 += int(floor((uf[1] - pow2mf) // -s1f[1]));
      # End of optimization.

      while True:
        # Check the condition on the radius.
        uf = [u2f[0] + i1 * s1f[0], u2f[1] + i1 * s1f[1]];
        if norm2(uf) > radius2:
          break;

        # Update the count with an additional candidate point. We only search
        # over positive i2, but if [i1, i2] is a candidate, then of course so is
        # [-i1, -i2], so unless i2 = 0 we count the candidate twice.
        if i2 != 0:
          count += 2;
        else:
          count += 1;

        # Check the candidate.
        if (not (i1 == i2 == 0)) and (0 < abs(uf[0]) < pow2mf) \
                                and (0 < abs(uf[1]) < pow2mf):

          # Compute r_tilde_candidate.
          r_tilde_candidate = abs(i1 * s1[1] + i2 * s2[1]);

          if (r_tilde_candidate >= 1) and (r_tilde_candidate < (2 ** m)):
            if r_tilde_candidate in filtered_r_tilde_candidates:
              success = True;

              if accept_multiple:
                return [filtered_r_tilde_candidates,
                        [success,
                        dismissed_reduced_r_tilde_candidates,
                        mu,
                        multiples]];
            else:
              # Use that mu is an r-multiple to reduce the r_tilde candidate.
              reduced_r_tilde_candidate = gcd(r_tilde_candidate, mu);

              if (reduced_r_tilde_candidate in \
                dismissed_reduced_r_tilde_candidates):
                # Dismiss the reduced candidate.
                if verbose:
                  print("Dismissing:", r_tilde_candidate);
              else:
                # The reduced candidate has not already been dismissed.
                if verbose:
                  print("Testing the candidate:", i1, i2, \
                    reduced_r_tilde_candidate, r_tilde_candidate);

                # Test the reduced candidate.
                if partial_exponentiation:
                  if None == x_basis:
                      x_basis = [x ** s1[1], x ** s2[1]];

                  z = (x_basis[0] ** i1) * (x_basis[1] ** i2);
                else:
                  z = x ** reduced_r_tilde_candidate;

                if z == 1:
                  success = True;

                  # Add r_tilde_candidate to the filtered r_tilde candidates.
                  filtered_r_tilde_candidates.add(r_tilde_candidate);

                  if accept_multiple:
                    return [filtered_r_tilde_candidates,
                            [success,
                            dismissed_reduced_r_tilde_candidates,
                            mu,
                            multiples]];

                  # We know that reduced_r_tilde_candidate * e is a multiple of
                  # r, so we may update mu to account for this fact:
                  mu = gcd(reduced_r_tilde_candidate * e, mu);
                else:
                  # Add reduced_r_tilde_candidate to the dismissed reduced
                  # candidates for r_tilde to avoid future exponentiations.
                  dismissed_reduced_r_tilde_candidates.\
                    add(reduced_r_tilde_candidate);

        if s1f[0] >= 0:
          if uf[0] >=  pow2mf:
            break; # There is no point in continuing.
        else:
          if uf[0] <= -pow2mf:
            break; # There is no point in continuing.

        if s1f[1] >= 0:
          if uf[1] >=  pow2mf:
            break; # There is no point in continuing.
        else:
          if uf[1] <= -pow2mf:
            break; # There is no point in continuing.

        i1 += 1;

      i1 = i1hat - 1;

      # Optimization: Jump ahead in i1.
      uf = [u2f[0] + i1 * s1f[0], u2f[1] + i1 * s1f[1]];

      if s1f[1] >= 0:
        if uf[1] <= -pow2mf:
          i1 -= int(floor((-uf[1] - pow2mf) // s1f[1]));
      else:
        if uf[1] >= pow2mf:
          i1 -= int(floor((uf[1] - pow2mf) // -s1f[1]));
      # End of optimization.

      while True:
        # Check the condition on the radius.
        uf = [u2f[0] + i1 * s1f[0], u2f[1] + i1 * s1f[1]];
        if norm2(uf) > radius2:
          break;

        # Update the count with an additional candidate point. We only search
        # over positive i2, but if [i1, i2] is a candidate, then of course so is
        # [-i1, -i2], so unless i2 = 0 we count the candidate twice.
        if i2 != 0:
          count += 2;
        else:
          count += 1;

        # Check the candidate.
        if (not (i1 == i2 == 0)) and (0 < abs(uf[0]) < pow2mf) \
                                and (0 < abs(uf[1]) < pow2mf):

          # Compute r_tilde_candidate.
          r_tilde_candidate = abs(i1 * s1[1] + i2 * s2[1]);

          if (r_tilde_candidate >= 1) and (r_tilde_candidate < (2 ** m)):
            if r_tilde_candidate in filtered_r_tilde_candidates:
              success = True;

              if accept_multiple:
                return [filtered_r_tilde_candidates,
                        [success,
                        dismissed_reduced_r_tilde_candidates,
                        mu,
                        multiples]];
            else:
              # Use that mu is an r-multiple to reduce the r_tilde candidate.
              reduced_r_tilde_candidate = gcd(r_tilde_candidate, mu);

              if (reduced_r_tilde_candidate in \
                dismissed_reduced_r_tilde_candidates):
                # Dismiss the reduced candidate.
                if verbose:
                  print("Dismissing:", r_tilde_candidate);
              else:
                # The reduced candidate has not already been dismissed.
                if verbose:
                  print("Testing the candidate:", i1, i2, \
                    reduced_r_tilde_candidate, r_tilde_candidate);

                # Test the reduced candidate.
                if partial_exponentiation:
                  if None == x_basis:
                      x_basis = [x ** s1[1], x ** s2[1]];

                  z = (x_basis[0] ** i1) * (x_basis[1] ** i2);
                else:
                  z = x ** reduced_r_tilde_candidate;

                if z == 1:
                  success = True;

                  # Add r_tilde_candidate to the filtered r_tilde candidates.
                  filtered_r_tilde_candidates.add(r_tilde_candidate);

                  if accept_multiple:
                    return [filtered_r_tilde_candidates,
                            [success,
                            dismissed_reduced_r_tilde_candidates,
                            mu,
                            multiples]];

                  # We know that reduced_r_tilde_candidate * e is a multiple of
                  # r, so we may update mu to account for this fact:
                  mu = gcd(reduced_r_tilde_candidate * e, mu);
                else:
                  # Add reduced_r_tilde_candidate to the dismissed reduced
                  # candidates for r_tilde to avoid future exponentiations.
                  dismissed_reduced_r_tilde_candidates.\
                    add(reduced_r_tilde_candidate);

        if s1f[0] <= 0:
          if uf[0] >=  pow2mf:
            break; # There is no point in continuing.
        else:
          if uf[0] <= -pow2mf:
            break; # There is no point in continuing.

        if s1f[1] <= 0:
          if uf[1] >=  pow2mf:
            break; # There is no point in continuing.
        else:
          if uf[1] <= -pow2mf:
            break; # There is no point in continuing.

        i1 -= 1;

      # Try next i2.
      i2 += 1;

      # Sanity check.
      if count >= B:
        raise Exception("Error: Enumerated more vectors than expected.");

    return [filtered_r_tilde_candidates,
            [success,
            dismissed_reduced_r_tilde_candidates,
            mu,
            multiples]];



""" @brief  A module for collecting candidates for r_tilde. """

class CandidateCollection:

  """ @brief  A class representing a collection of candidates for r_tilde, that
              is kept reduced in the sense that if r_tilde is in the collection,
              then so is c * r_tilde, for c any positive integer.

      The idea is to only add candidates for r_tilde that meet the requirement
      that g^(e * r_tilde) = 1 to the collection, for e a product of cm-smooth
      prime powers. If r_tilde meets this requirement, then of course so does
      c * r_tilde, so it makes sense to keep the collection reduced.

      In the internal set of candidates that represent the collection, only the
      minimum set of candidates for r_tilde required to represent the collection
      are stored. The collection furthermore has features for testing if a
      candidate is already in the collection. This is useful when searching for
      r_tilde, as it reduces the number of exponentiations required, and the
      amount of storage needed to represent candidates, and so forth. """

  def __init__(self):

    """ @brief  Initializes the collection of candidates for r_tilde. """

    self.candidates = set();

  def add(self, candidate):

    """ @brief  Adds a candidate for r_tilde to this collection.

        @param candidate  The candidate for r_tilde.

        @return True if the collection was modified, False otherwise. """

    if candidate in self:
      # The candidate is already in the collection.
      return False;

    for x in self.candidates.copy():
      if (x % candidate) == 0:
        self.candidates.remove(x);

    self.candidates.add(candidate);

    return True;

  def __contains__(self, candidate):

    """ @brief  Checks if a candidate for r_tilde is in this collection.

        @param candidate  The candidate for r_tilde.

        @return True if the candidate is in the collection, False otherwise. """

    if candidate in self.candidates:
      return True;

    for x in self.candidates:
      if (candidate % x) == 0:
        return True;

    return False;

  def __len__(self):

    """ @brief  Returns the number of candidates for r_tilde that represent this
                collection.

        @remark   Note that the collection is kept reduced, so the number of
                  candidates for r_tilde that represent the collection may be
                  fewer than the number of candidates added to the collection.

        @return   The number of candidates for r_tilde that represent the
                  collection. """

    return len(self.candidates);

  def __iter__(self):

    """ @brief  Returns an iterator for the candidates for r_tilde that
                represent this collection.

        @remark   Note that the collection is kept reduced, so the number of
                  candidates for r_tilde that represent the collection may be
                  fewer than the number of candidates added to the collection.

        @return   An iterator for the candidates for r_tilde that represent the
                  collection. """

    return self.candidates.__iter__();

  def __str__(self):

    """ @brief  Returns a string representation of this collection.

        @return   A string representation of this collection. """

    return str(self.candidates);

  def __repr__(self):

    """ @brief  Returns a string representation of this collection.

        @return   A string representation of this collection. """

    return str(self);


""" @brief  A module for sampling a frequency pair (j, k) heuristically from the
            distribution induced by the quantum part of Shor's algorithm for
            finding a discrete logarithm d in a group of known order r, or from
            the distribution induced by the quantum part of Ekerå's variation
            thereof, depending on how parameters are selected. """

from ..sampling import sample_j_k_given_d_r_heuristic as \
  sample_j_k_given_d_r_heuristic_general;

from ...utils.timeout import Timeout;

from ..sampling import B_DEFAULT_DELTA;
from ..sampling import B_DEFAULT_ETA;
from ..sampling import DEFAULT_INTEGRATION_STEPS;

from gmpy2 import mpz;

def sample_j_k_given_d_r_heuristic(
  d : int | mpz,
  r : int | mpz,
  m : int,
  sigma : int,
  l : int,
  B_DELTA : int = B_DEFAULT_DELTA,
  B_ETA : int = B_DEFAULT_ETA,
  integration_steps : int = DEFAULT_INTEGRATION_STEPS,
  timeout : int | None | Timeout = None,
  verbose : bool = False,
  extended_result : bool = False) -> list[int | mpz] | None:

  """ @brief  Samples a frequency pair (j, k) heuristically from the
              distribution induced by the quantum part of Shor's algorithm for
              finding a discrete logarithm d in a group of known order r, or
              from the distribution induced by the quantum part of Ekerå's
              variation thereof, depending on how parameters are selected.

      The sampling procedure is described in Sect. 5 of [E19p].

      [E19p] Ekerå, M.: "Revisiting Shor's quantum algorithm for computing
                         general discrete logarithms".
                        ArXiv 1905.09084v4 (2024).

      @param d  The discrete logarithm d in [1, r).

      @param r  The order r.

      @param m  A positive integer m such that r < 2^m.

      @param sigma  A non-negative integer sigma such that m + sigma is the
                    length of the first control register in the quantum part of
                    the algorithm.

      @param l  A positive integer l such that l is the length of the second
                control register in the quantum part of the algorithm.

      @param B_DELTA  A parameter that upper-bounds the offset from the optimal
                      frequency k0(j) when sampling k given j and eta.

      @param B_ETA  A parameter that upper-bounds eta when sampling j and eta.

      @param integration_steps  The number of steps to perform when integrating
                                the probability distribution.

      @param timeout  A timeout after which a TimeoutError will be raised and
                      the sampling procedure aborted.

                      The timeout may be represented as an integer specifying
                      the timeout in seconds, or as an instance of the Timeout
                      class. May be set to None, as is the default, in which
                      case no timeout is enforced.

      @param verbose  A flag that may be set to True to print intermediary
                      results when sampling.

      @param extended_result  A flag that may be set to True to not only return
                              the frequency pair (j, k) as a list [j, k], but
                              [[j, k], [k0(j), offset, eta]].

      @return   The frequency pair [j, k] sampled if the extended_result flag is
                set to False, or [[j, k], [k0(j), offset, eta]] if the
                extended_result flag is set to True, or None if sampling failed
                because the upper bound on the offset from the optimal frequency
                k0(j) or on eta were reached. """

  return sample_j_k_given_d_r_heuristic_general(
           d = d,
           r = r,
           m = m,
           sigma = sigma,
           l = l,
           B_DELTA = B_DELTA,
           B_DEFAULT_ETA = B_ETA,
           integration_steps = integration_steps,
           verbose = verbose,
           timeout = timeout,
           extended_result = extended_result);



""" @brief  A module for enumerating vectors in lattices. """

from .gram_schmidt import gram_schmidt;

from ..matrices import dimensions;
from ..matrices import solve_left;

from ..vectors import norm2;

from ...utils.timeout import Timeout;

from typing import Generator;

from gmpy2 import mpz;
from gmpy2 import mpq;
from gmpy2 import mpfr;

def enumerate(
  B : list[list[int | mpz]],
  R2 : int | mpz | mpq | mpfr,
  t : list[int | mpz | mpq | float | mpfr] | None = None,
  timeout : int | None | Timeout = None,
  gs : list[list[list[int | mpz | mpq | mpfr]]] | None = None,
  precision : int | None = None) -> \
    Generator[list[int | mpz], None, None]:

  """ @brief  Enumerates all vectors in the lattice L generated by the rows
              of the basis B that are within distance R of the vector t.

      The n x d integer basis matrix B is represented as as list of lists
      [b_1, ..., b_n], for b_i = [b_i1, ..., b_id] for i = 1, ..., n the n row
      vectors that make up B where b_ij is of type int or mpz.

      This function currently requires B to be square and to have full row rank
      since it calls solve_left(B, t) which takes the inverse of B to solve.
      These requirement may be relaxed in the future.

      @param B  The n x d basis matrix B = [b_1, ..., b_n], where
                b_i = [b_i1, ..., b_id] for i = 1, ..., n represent the n row
                vectors that make up the basis.

      @param R2  The square of the enumeration radius R.

      @param t  The vector t = [t_1, ..., t_d] in span(L). May be set to None,
                in which case t = [0, ..., 0].

      @param timeout  A timeout in seconds after which a TimeoutError will be
                      raised and the enumeration aborted. The timeout is
                      represented as integer or as an instance of Timeout. May
                      be set to None, as is the default, in which case no
                      timeout is enforced.

                      Note this function incrementally yields the vectors that
                      are found during the enumeration. The timeout specified
                      here is respect to the time elapsed since enumerate() was
                      first called. It hence includes any time spent processing
                      the results yielded by this function until control returns
                      to this function.

      @param gs   The list [Bs, M], where Bs is the Gram–Schmidt orthogonalized
                  basis for B, and M is the associated matrix of Gram–Schmidt
                  projection factors, as returned by calling

                     [Bs, M] = gram_schmidt(B, precision = precision),

                  or None, in which case this function will make the above call.

                  If you plan to perform several enumerations in the same
                  lattice, then time may be saved by not re-computing Bs and M
                  for each call.

                  Note that the basis B is typically LLL-reduced by calling
                  lll() before calling this function. The lll() function can
                  return not only B but also Bs and M (since Bs and M are
                  incrementally computed as a part of the LLL reduction process)
                  allowing you to directly pass them along to this function.

      @param precision  The precision to use when computing the Gram–Schmidt
                        projection factors in M. May be set to None, in which
                        case the projection factors are represented as exact
                        quotients.

                        Note that this parameter only has an effect if gs is
                        set to None as the Gram–Schmidt orthogonalized basis Bs
                        and the associated matrix M of Gram–Schmidt projection
                        factors are otherwise pre-computed.

      return  A list [x_1, ..., x_m] of the m vectors found, where each vector
              x_i = [x_i1, ..., x_id] for 1, ..., m. This function yields one
              vector at a time allowing you to enumerate over the vectors. """

  # Initial setup.
  timeout = Timeout.parse(timeout);
  timeout.check();

  (n, d) = dimensions(B);

  if None == gs:
    [Bs, M] = gram_schmidt(B, precision = precision);
  else:
    [Bs, M] = gs;

  timeout.check();

  # 1
  if None == t:
    t = [0 for _ in range(d)];
    ct = [0 for _ in range(n)];
  else:
    ct = solve_left(B, t);

  cx = [0 for _ in range(n)];

  # 2
  def enumerate_inner(cx, ct, k):

    # Check the timeout.
    timeout.check();

    # 2.1
    if k == 0:

      # 2.1.1
      x = [sum([cx[i] * B[i][j] for i in range(n)]) for j in range(d)];
      yield x;

      return;

    # 2.2
    cx[k - 1] = 0;

    # 2.3
    a = ct[k - 1] - sum([M[i - 1][k - 1] * (cx[i - 1] - ct[i - 1]) \
                         for i in range(k + 1, n + 1)]);

    pi_k_1 = 0;
    for j in range(k + 1, n + 1):
      term = ((cx[j - 1] - ct[j - 1]) + \
                sum([M[i - 1][j - 1] * (cx[i - 1] - ct[i - 1]) \
                       for i in range(j + 1, n + 1)])) ** 2;
      term *= norm2(Bs[j - 1]);
      pi_k_1 += term;

    B2 = (R2 - pi_k_1) / norm2(Bs[k - 1]);
    if B2 < 0:
      return; # Return nothing.

    cx[k - 1] = round(a)
    tmp = abs(cx[k - 1] - a);
    if tmp * tmp <= B2:
      yield from enumerate_inner(cx.copy(), ct.copy(), k - 1);

    o = 0;
    proceed = True;

    while proceed:

      # Check the timeout.
      timeout.check();

      o = o + 1;
      proceed = False;

      cx[k - 1] = round(a) + o;
      tmp = abs(cx[k - 1] - a);
      if tmp * tmp <= B2:
        yield from enumerate_inner(cx.copy(), ct.copy(), k - 1);
        proceed = True;

      cx[k - 1] = round(a) - o;
      tmp = abs(cx[k - 1] - a);
      if tmp * tmp <= B2:
        yield from enumerate_inner(cx.copy(), ct.copy(), k - 1);
        proceed = True;

  # 3
  yield from enumerate_inner(cx, ct, n);



""" @brief  A module for vectors. """

from typing import Any;

def norm2(u):

  """ @brief  Returns the square norm of the n-dimensional vector u.

      @param u  The n-dimensional vector u = [u_1, ..., u_n].

      @return   The square norm of u. """

  return sum([ui * ui for ui in u]);